from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from qcportal.auth.models import UserInfo, GroupInfo, is_valid_password
from qcportal.exceptions import (
    UserManagementError,
    AuthenticationFailure,
    InvalidPasswordError,
    InvalidUsernameError,
)

if TYPE_CHECKING:
    from qcfractal.db_socket import SQLAlchemySocket

invalid_usernames = ["\x00", "ab\x00cd", "a user", ""]
invalid_passwords = ["\x00", "abcd\x00efgh", "abcd", "1", ""]


def test_user_socket_add_get(storage_socket: SQLAlchemySocket):
    storage_socket.groups.add(GroupInfo(groupname="group1"))
    storage_socket.groups.add(GroupInfo(groupname="group2"))

    uinfo = UserInfo(
        username="george",
        role="read",
        groups=["group1", "group2"],
        enabled=True,
        fullname="Test user",
        email="george@example.com",
        organization="My Org",
    )
    pw = storage_socket.users.add(uinfo, password="oldpw123")
    assert pw == "oldpw123"

    # Do we get the same data back?
    # The initial userinfo doesn't contain the id
    uinfo2 = storage_socket.users.get("george")
    assert "password" not in uinfo2
    uinfo2.pop("id")
    assert UserInfo(**uinfo2) == uinfo

    # Get by id
    uinfo2 = storage_socket.users.get("george")
    uinfo3 = storage_socket.users.get(uinfo2["id"])
    assert uinfo2 == uinfo3


def test_user_socket_add_duplicate(storage_socket: SQLAlchemySocket):
    uinfo = UserInfo(
        username="george",
        role="read",
        enabled=True,
    )
    storage_socket.users.add(uinfo, password="oldpw123")

    # Duplicate should result in an exception
    uinfo2 = UserInfo(username="george", role="read", enabled=True)
    with pytest.raises(UserManagementError, match=r"User.*already exists"):
        storage_socket.users.add(uinfo2, "newpw123")


def test_user_socket_add_with_id(storage_socket: SQLAlchemySocket):
    # Should not be able to add a user with the id set
    uinfo = UserInfo(
        id=123,
        username="george",
        role="read",
        enabled=True,
    )

    with pytest.raises(UserManagementError, match=r"id was given as part"):
        storage_socket.users.add(uinfo, password="oldpw123")


def test_user_socket_list(storage_socket: SQLAlchemySocket):
    all_users = []
    for i in range(20):
        uinfo = UserInfo(
            username=f"george_{i}",
            role="read",
            enabled=bool(i % 2),
            fullname=f"Test user_{i}",
            email=f"george{i}@example.com",
            organization=f"My Org {i}",
        )
        storage_socket.users.add(uinfo)
        all_users.append(uinfo)

    user_lst = storage_socket.users.list()
    user_lst_model = [UserInfo(**x) for x in user_lst]

    # Sort both lists by username
    all_users = sorted(all_users, key=lambda x: x.username)
    user_lst_model = sorted(user_lst_model, key=lambda x: x.username)

    d1 = [x.dict(exclude={"id"}) for x in all_users]
    d2 = [x.dict(exclude={"id"}) for x in user_lst_model]
    assert d1 == d2


def test_user_socket_delete(storage_socket: SQLAlchemySocket):
    uinfo1 = UserInfo(
        username="george",
        role="read",
        enabled=True,
    )
    uinfo2 = UserInfo(
        username="bill",
        role="read",
        enabled=True,
    )
    storage_socket.users.add(uinfo1)
    storage_socket.users.add(uinfo2)

    uid1 = storage_socket.users.get("george")["id"]
    uid2 = storage_socket.users.get("bill")["id"]

    # Raises exception on error
    storage_socket.users.delete("george")
    storage_socket.users.delete(uid2)

    with pytest.raises(UserManagementError, match=r"User.*not found"):
        storage_socket.users.get("george")
    with pytest.raises(UserManagementError, match=r"User.*not found"):
        storage_socket.users.get(uid1)
    with pytest.raises(UserManagementError, match=r"User.*not found"):
        storage_socket.users.get("bill")
    with pytest.raises(UserManagementError, match=r"User.*not found"):
        storage_socket.users.get(uid2)


def test_user_socket_use_unknown_user(storage_socket: SQLAlchemySocket):
    uinfo = UserInfo(
        username="george",
        role="read",
        enabled=True,
    )
    storage_socket.users.add(uinfo)

    with pytest.raises(UserManagementError, match=r"User.*not found"):
        storage_socket.users.get("geoff")

    with pytest.raises(AuthenticationFailure, match=r"Incorrect username or password"):
        storage_socket.users.authenticate("geoff", "a password")

    with pytest.raises(UserManagementError, match=r"User.*not found"):
        uinfo = UserInfo(id=1234, username="geoff", role="read", enabled=True)
        storage_socket.users.modify(uinfo, False)

    with pytest.raises(UserManagementError, match=r"User.*not found"):
        storage_socket.users.change_password("geoff", "a password")

    with pytest.raises(UserManagementError, match=r"User.*not found"):
        storage_socket.users.change_password("geoff", None)

    with pytest.raises(UserManagementError, match=r"User.*not found"):
        storage_socket.users.delete("geoff")


def test_user_socket_verify_password(storage_socket: SQLAlchemySocket):
    for idx, password in enumerate(["simple", "ABC 1234", "ÃØ©þꝎꟇ"]):
        username = f"george_{idx}"
        uinfo = UserInfo(
            username=username,
            role="read",
            enabled=True,
        )

        add_pw = storage_socket.users.add(uinfo, password=password)
        assert add_pw == password
        storage_socket.users.authenticate(username, add_pw)

        for guess in ["Simple", "ABC%1234", "ÃØ©þꝎB"]:
            with pytest.raises(AuthenticationFailure):
                storage_socket.users.authenticate(username, guess)


def test_user_socket_verify_user_disabled(storage_socket: SQLAlchemySocket):
    uinfo = UserInfo(
        username="george",
        role="read",
        enabled=True,
    )

    gen_pw = storage_socket.users.add(uinfo)

    uinfo2 = storage_socket.users.authenticate("george", gen_pw)

    uinfo2.enabled = False
    storage_socket.users.modify(uinfo2, as_admin=True)

    with pytest.raises(AuthenticationFailure):
        storage_socket.users.authenticate("george", gen_pw)


def test_user_socket_change_password(storage_socket: SQLAlchemySocket):
    uinfo = UserInfo(
        username="george",
        role="read",
        enabled=True,
    )

    old_pw = storage_socket.users.add(uinfo, "oldpw123")
    assert old_pw == "oldpw123"

    storage_socket.users.authenticate("george", "oldpw123")

    # update password...
    storage_socket.users.change_password("george", password="newpw123")

    # Raises exception on failure
    storage_socket.users.authenticate("george", "newpw123")

    with pytest.raises(AuthenticationFailure):
        storage_socket.users.authenticate("george", "oldpw123")


def test_user_socket_password_generation(storage_socket: SQLAlchemySocket):
    uinfo = UserInfo(
        username="george",
        role="read",
        enabled=True,
    )

    gen_pw = storage_socket.users.add(uinfo)
    storage_socket.users.authenticate("george", gen_pw)
    is_valid_password(gen_pw)
    storage_socket.users.authenticate("george", gen_pw)

    gen_pw_2 = storage_socket.users.change_password("george", None)
    storage_socket.users.authenticate("george", gen_pw_2)
    is_valid_password(gen_pw)

    with pytest.raises(AuthenticationFailure):
        storage_socket.users.authenticate("george", gen_pw)


@pytest.mark.parametrize("as_admin", [True, False])
def test_user_socket_no_modify_username(storage_socket: SQLAlchemySocket, as_admin: bool):
    uinfo = UserInfo(
        username="george",
        role="read",
        enabled=False,
        fullname="Test user",
        email="george@example.com",
        organization="My Org",
    )
    storage_socket.users.add(uinfo)
    uid = storage_socket.users.get("george")["id"]

    uinfo2 = UserInfo(
        id=uid, username="george2", role="admin", fullname="Test user 2", email="test@example.com", enabled=True
    )

    with pytest.raises(UserManagementError, match=r"Cannot change"):
        uinfo3 = storage_socket.users.modify(uinfo2, as_admin=as_admin)


@pytest.mark.parametrize("as_admin", [True, False])
def test_user_socket_modify(storage_socket: SQLAlchemySocket, as_admin: bool):
    # If as_admin == True for user.modify(), then all fields can be modified
    # Otherwise, some fields will always stay the same (enabled, role)

    uinfo = UserInfo(
        username="george",
        role="read",
        enabled=False,
        fullname="Test user",
        email="george@example.com",
        organization="My Org",
    )
    storage_socket.users.add(uinfo)
    uid = storage_socket.users.get("george")["id"]

    uinfo2 = UserInfo(
        id=uid, username="george", role="admin", fullname="Test user 2", email="test@example.com", enabled=True
    )

    # Modify should return the same this as get
    uinfo3 = storage_socket.users.modify(uinfo2, as_admin=as_admin)
    uinfo4 = storage_socket.users.get("george")

    assert uinfo3 == UserInfo(**uinfo4)

    if as_admin is True:
        assert uinfo2 == UserInfo(**uinfo3)
    else:
        # Stayed the same
        assert uinfo3["enabled"] == uinfo.enabled
        assert uinfo3["role"] == uinfo.role

        # Can be modified
        assert uinfo3["fullname"] == uinfo2.fullname
        assert uinfo3["email"] == uinfo2.email
        assert uinfo3["organization"] == uinfo2.organization


def test_user_socket_use_invalid_username(storage_socket: SQLAlchemySocket):
    for username in invalid_usernames:
        # Normally, UserInfo prevents bad usernames. But the socket also checks, as a last resort
        # So we have to bypass the UserInfo check with construct()
        uinfo = UserInfo.construct(
            username=username,
            role="read",
            enabled=True,
        )

        with pytest.raises(InvalidUsernameError):
            storage_socket.users.add(uinfo, "password123")

        with pytest.raises(InvalidUsernameError):
            storage_socket.users.get(username)

        with pytest.raises(InvalidUsernameError):
            storage_socket.users.authenticate(username, "a_password")

        with pytest.raises(InvalidUsernameError):
            storage_socket.users.change_password(username, "a_password")

        with pytest.raises(InvalidUsernameError):
            storage_socket.users.change_password(username, None)

        with pytest.raises(InvalidUsernameError):
            storage_socket.users.delete(username)

    # Numeric usernames not allowed for some operations
    uinfo2 = UserInfo.construct(
        username="123456789",
        role="read",
        enabled=True,
    )

    with pytest.raises(InvalidUsernameError):
        storage_socket.users.add(uinfo2, "password123")

    with pytest.raises(InvalidUsernameError):
        storage_socket.users.authenticate("123456789", "a_password")


def test_user_socket_use_invalid_password(storage_socket: SQLAlchemySocket):
    for idx, password in enumerate(invalid_passwords):
        username = f"george_{idx}"

        uinfo = UserInfo.construct(
            username=username,
            role="read",
            enabled=True,
        )
        with pytest.raises(InvalidPasswordError):
            storage_socket.users.add(uinfo, password)

        #  Add for real now
        storage_socket.users.add(uinfo, "good_password")
        uid = storage_socket.users.get(username)["id"]

        with pytest.raises(InvalidPasswordError):
            storage_socket.users.change_password(username, password)

        with pytest.raises(InvalidPasswordError):
            storage_socket.users.change_password(uid, password)

        with pytest.raises(InvalidPasswordError):
            storage_socket.users.authenticate(username, password)



from __future__ import annotations

import threading
import weakref
from typing import Optional

from sqlalchemy import update
from sqlalchemy.orm import Session

from qcfractal.components.internal_jobs.db_models import InternalJobORM
from qcportal.internal_jobs.models import InternalJobStatusEnum
from qcportal.utils import now_at_utc


class CancelledJobException(Exception):
    pass


class JobRunnerStoppingException(Exception):
    pass


class JobProgress:
    """
    Functor for updating progress and cancelling internal jobs
    """

    def __init__(self, job_id: int, runner_uuid: str, session: Session, update_frequency: int, end_event):
        self._update_frequency = update_frequency
        self._job_id = job_id
        self._runner_uuid = runner_uuid
        self._stmt = (
            update(InternalJobORM)
            .where(InternalJobORM.id == self._job_id)
            .returning(InternalJobORM.status, InternalJobORM.runner_uuid)
        )
        self._progress = 0
        self._description = None

        self._cancelled = False
        self._runner_ending = False
        self._deleted = False

        self._end_event = end_event

        # A thread for periodically updating
        # _th_cancel is set when we want the updating thread to end
        self._th_cancel = threading.Event()
        self._th = threading.Thread(target=self._update_thread, args=(session, self._th_cancel))

        # Start the updating thread
        self._th.start()

        # Create the finalizer that will close the updating thread
        self._finalizer = weakref.finalize(self, self._stop_thread, self._th_cancel, self._th)

    def _update_thread(self, session: Session, end_thread: threading.Event):
        while True:
            # Update progress
            stmt = self._stmt.values(
                progress=self._progress, progress_description=self._description, last_updated=now_at_utc()
            )
            ret = session.execute(stmt).one_or_none()
            session.commit()

            if ret is None:
                # Job was deleted
                self._cancelled = True
                self._deleted = True
            elif ret[0] != InternalJobStatusEnum.running:
                # Job was cancelled or something
                self._cancelled = True
            elif ret[1] != self._runner_uuid:
                # Job was stolen from us?
                self._cancelled = True

            # Are we ending/cancelling because the runner is stopping/closing?
            if self._end_event.is_set():
                self._runner_ending = True

            cancel = end_thread.wait(self._update_frequency)
            if cancel is True:
                break

    @staticmethod
    def _stop_thread(cancel_event: threading.Event, thread: threading.Thread):
        ####################################################################################
        # This is written as a class method so that it can be called by a weakref finalizer
        ####################################################################################

        cancel_event.set()
        thread.join()

    def stop(self):
        self._finalizer()

    def update_progress(self, progress: int, description: Optional[str] = None):
        self._progress = progress
        self._description = description

    @property
    def cancelled(self) -> bool:
        return self._cancelled or self._deleted

    @property
    def runner_ending(self) -> bool:
        return self._runner_ending

    @property
    def deleted(self) -> bool:
        return self._deleted

    def raise_if_cancelled(self):
        if self._cancelled or self._deleted:
            raise CancelledJobException("Job was cancelled or deleted")

        if self._runner_ending:
            raise JobRunnerStoppingException("Job runner is stopping/ending")


def raise_if_cancelled(job_progress: Optional[JobProgress]):
    """
    Raises a CancelledJobExcepion if job_progress exists and is in a cancelled or deleted state
    """

    if job_progress is not None:
        job_progress.raise_if_cancelled()



"""use views for dataset records

Revision ID: 5ae71b845526
Revises: ee68b4a1c346
Create Date: 2025-04-15 19:43:41.069422

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "5ae71b845526"
down_revision = "ee68b4a1c346"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute(
        sa.text(
            """
                create or replace view "public"."dataset_direct_records_view" as  SELECT base_dataset.id AS dataset_id,
                    singlepoint_dataset_record.entry_name,
                    singlepoint_dataset_record.specification_name,
                    singlepoint_dataset_record.record_id
                   FROM (base_dataset
                     JOIN singlepoint_dataset_record ON ((base_dataset.id = singlepoint_dataset_record.dataset_id)))
                  WHERE ((base_dataset.dataset_type)::text = 'singlepoint'::text)
                UNION
                 SELECT base_dataset.id AS dataset_id,
                    optimization_dataset_record.entry_name,
                    optimization_dataset_record.specification_name,
                    optimization_dataset_record.record_id
                   FROM (base_dataset
                     JOIN optimization_dataset_record ON ((base_dataset.id = optimization_dataset_record.dataset_id)))
                  WHERE ((base_dataset.dataset_type)::text = 'optimization'::text)
                UNION
                 SELECT base_dataset.id AS dataset_id,
                    gridoptimization_dataset_record.entry_name,
                    gridoptimization_dataset_record.specification_name,
                    gridoptimization_dataset_record.record_id
                   FROM (base_dataset
                     JOIN gridoptimization_dataset_record ON ((base_dataset.id = gridoptimization_dataset_record.dataset_id)))
                  WHERE ((base_dataset.dataset_type)::text = 'gridoptimization'::text)
                UNION
                 SELECT base_dataset.id AS dataset_id,
                    torsiondrive_dataset_record.entry_name,
                    torsiondrive_dataset_record.specification_name,
                    torsiondrive_dataset_record.record_id
                   FROM (base_dataset
                     JOIN torsiondrive_dataset_record ON ((base_dataset.id = torsiondrive_dataset_record.dataset_id)))
                  WHERE ((base_dataset.dataset_type)::text = 'torsiondrive'::text)
                UNION
                 SELECT base_dataset.id AS dataset_id,
                    manybody_dataset_record.entry_name,
                    manybody_dataset_record.specification_name,
                    manybody_dataset_record.record_id
                   FROM (base_dataset
                     JOIN manybody_dataset_record ON ((base_dataset.id = manybody_dataset_record.dataset_id)))
                  WHERE ((base_dataset.dataset_type)::text = 'manybody'::text)
                UNION
                 SELECT base_dataset.id AS dataset_id,
                    reaction_dataset_record.entry_name,
                    reaction_dataset_record.specification_name,
                    reaction_dataset_record.record_id
                   FROM (base_dataset
                     JOIN reaction_dataset_record ON ((base_dataset.id = reaction_dataset_record.dataset_id)))
                  WHERE ((base_dataset.dataset_type)::text = 'reaction'::text)
                UNION
                 SELECT base_dataset.id AS dataset_id,
                    neb_dataset_record.entry_name,
                    neb_dataset_record.specification_name,
                    neb_dataset_record.record_id
                   FROM (base_dataset
                     JOIN neb_dataset_record ON ((base_dataset.id = neb_dataset_record.dataset_id)))
                  WHERE ((base_dataset.dataset_type)::text = 'neb'::text);
    """
        )
    )

    op.execute(
        sa.text(
            """
              create or replace view "public"."dataset_records_view" as  SELECT dataset_direct_records_view.dataset_id,
                  dataset_direct_records_view.record_id
                 FROM dataset_direct_records_view
              UNION
               SELECT dataset_direct_records_view.dataset_id,
                  record_children_view.child_id AS record_id
                 FROM (dataset_direct_records_view
                   JOIN record_children_view ON ((dataset_direct_records_view.record_id = record_children_view.parent_id)));
    """
        )
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute(sa.text("DROP VIEW dataset_records_view;"))
    op.execute(sa.text("DROP VIEW dataset_direct_records_view;"))
    # ### end Alembic commands ###



"""Make DB more timezone aware

Revision ID: de44b1b34097
Revises: 13cb230def11
Create Date: 2023-11-27 12:14:40.915673

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "de44b1b34097"
down_revision = "13cb230def11"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "access_log",
        "timestamp",
        type_=sa.TIMESTAMP(timezone=True),
        postgresql_using="timestamp at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "base_record",
        "created_on",
        type_=sa.TIMESTAMP(timezone=True),
        postgresql_using="created_on at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "base_record",
        "modified_on",
        type_=sa.TIMESTAMP(timezone=True),
        postgresql_using="modified_on at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "compute_manager",
        "created_on",
        type_=sa.TIMESTAMP(timezone=True),
        postgresql_using="created_on at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "compute_manager",
        "modified_on",
        type_=sa.TIMESTAMP(timezone=True),
        postgresql_using="modified_on at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "compute_manager_log",
        "timestamp",
        type_=sa.TIMESTAMP(timezone=True),
        postgresql_using="timestamp at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "internal_error_log",
        "error_date",
        type_=sa.TIMESTAMP(timezone=True),
        postgresql_using="error_date at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "internal_jobs",
        "added_date",
        type_=sa.TIMESTAMP(timezone=True),
        postgresql_using="added_date at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "internal_jobs",
        "scheduled_date",
        type_=sa.TIMESTAMP(timezone=True),
        postgresql_using="scheduled_date at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "internal_jobs",
        "started_date",
        type_=sa.TIMESTAMP(timezone=True),
        postgresql_using="started_date at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "internal_jobs",
        "last_updated",
        type_=sa.TIMESTAMP(timezone=True),
        postgresql_using="last_updated at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "internal_jobs",
        "ended_date",
        type_=sa.TIMESTAMP(timezone=True),
        postgresql_using="ended_date at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "record_comment",
        "timestamp",
        type_=sa.TIMESTAMP(timezone=True),
        postgresql_using="timestamp at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "record_compute_history",
        "modified_on",
        type_=sa.TIMESTAMP(timezone=True),
        postgresql_using="modified_on at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "record_info_backup",
        "modified_on",
        type_=sa.TIMESTAMP(timezone=True),
        postgresql_using="modified_on at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "server_stats_log",
        "timestamp",
        type_=sa.TIMESTAMP(timezone=True),
        postgresql_using="timestamp at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "server_stats_metadata",
        "date_value",
        type_=sa.TIMESTAMP(timezone=True),
        postgresql_using="date_value at time zone 'Etc/UTC'",
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "access_log",
        "timestamp",
        type_=sa.TIMESTAMP(timezone=False),
        postgresql_using="timestamp at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "base_record",
        "created_on",
        type_=sa.TIMESTAMP(timezone=False),
        postgresql_using="created_on at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "base_record",
        "modified_on",
        type_=sa.TIMESTAMP(timezone=False),
        postgresql_using="modified_on at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "compute_manager",
        "created_on",
        type_=sa.TIMESTAMP(timezone=False),
        postgresql_using="created_on at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "compute_manager",
        "modified_on",
        type_=sa.TIMESTAMP(timezone=False),
        postgresql_using="modified_on at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "compute_manager_log",
        "timestamp",
        type_=sa.TIMESTAMP(timezone=False),
        postgresql_using="timestamp at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "internal_error_log",
        "error_date",
        type_=sa.TIMESTAMP(timezone=False),
        postgresql_using="error_date at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "internal_jobs",
        "added_date",
        type_=sa.TIMESTAMP(timezone=False),
        postgresql_using="added_date at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "internal_jobs",
        "scheduled_date",
        type_=sa.TIMESTAMP(timezone=False),
        postgresql_using="scheduled_date at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "internal_jobs",
        "started_date",
        type_=sa.TIMESTAMP(timezone=False),
        postgresql_using="started_date at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "internal_jobs",
        "last_updated",
        type_=sa.TIMESTAMP(timezone=False),
        postgresql_using="last_updated at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "internal_jobs",
        "ended_date",
        type_=sa.TIMESTAMP(timezone=False),
        postgresql_using="ended_date at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "record_comment",
        "timestamp",
        type_=sa.TIMESTAMP(timezone=False),
        postgresql_using="timestamp at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "record_compute_history",
        "modified_on",
        type_=sa.TIMESTAMP(timezone=False),
        postgresql_using="modified_on at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "record_info_backup",
        "modified_on",
        type_=sa.TIMESTAMP(timezone=False),
        postgresql_using="modified_on at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "server_stats_log",
        "timestamp",
        type_=sa.TIMESTAMP(timezone=False),
        postgresql_using="timestamp at time zone 'Etc/UTC'",
    )
    op.alter_column(
        "server_stats_metadata",
        "date_value",
        type_=sa.TIMESTAMP(timezone=False),
        postgresql_using="date_value at time zone 'Etc/UTC'",
    )



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from qcelemental.models import AtomicResult as QCEl_AtomicResult
from sqlalchemy import select, or_
from sqlalchemy.orm import lazyload, joinedload, defer, undefer, defaultload, load_only, selectinload

from qcfractal.db_socket.helpers import insert_general
from qcportal.compression import CompressionEnum, compress
from qcportal.exceptions import MissingDataError
from qcportal.metadata_models import InsertMetadata, InsertCountsMetadata
from qcportal.molecules import Molecule
from qcportal.record_models import PriorityEnum, RecordStatusEnum
from qcportal.serialization import convert_numpy_recursive
from qcportal.singlepoint import (
    QCSpecification,
    WavefunctionProperties,
    SinglepointQueryFilters,
    SinglepointInput,
    SinglepointMultiInput,
    SinglepointRecord,
)
from qcportal.utils import hash_dict, is_included
from .record_db_models import QCSpecificationORM, SinglepointRecordORM, WavefunctionORM
from ..base_record_socket import BaseRecordSocket

if TYPE_CHECKING:
    from sqlalchemy.orm.session import Session
    from qcfractal.db_socket.socket import SQLAlchemySocket
    from typing import List, Dict, Tuple, Optional, Sequence, Any, Union

# Meaningless, but unique to singlepoints
singlepoint_insert_lock_id = 14000


class SinglepointRecordSocket(BaseRecordSocket):
    """
    Socket for handling singlepoint computations
    """

    # Used by the base class
    record_orm = SinglepointRecordORM
    record_input_type = SinglepointInput
    record_multi_input_type = SinglepointMultiInput

    def __init__(self, root_socket: SQLAlchemySocket):
        BaseRecordSocket.__init__(self, root_socket)
        self._logger = logging.getLogger(__name__)

    def generate_task_specifications(self, session: Session, record_ids: Sequence[int]) -> List[Dict[str, Any]]:
        stmt = select(SinglepointRecordORM).filter(SinglepointRecordORM.id.in_(record_ids))
        stmt = stmt.options(load_only(SinglepointRecordORM.id, SinglepointRecordORM.extras))
        stmt = stmt.options(
            lazyload("*"), joinedload(SinglepointRecordORM.molecule), selectinload(SinglepointRecordORM.specification)
        )

        record_orms = session.execute(stmt).scalars().all()

        task_specs = {}

        for record_orm in record_orms:
            specification = record_orm.specification
            molecule = record_orm.molecule.model_dict()

            model = {"method": specification.method}

            # Empty basis string should be None in the model
            if specification.basis:
                model["basis"] = specification.basis
            else:
                model["basis"] = None

            qcschema_input = dict(
                schema_name="qcschema_input",
                schema_version=1,
                id=str(record_orm.id),  # str for compatibility
                driver=specification.driver,
                model=model,
                molecule=convert_numpy_recursive(molecule, flatten=True),  # TODO - remove after all data is converted
                keywords=specification.keywords,
                protocols=specification.protocols,
                extras=record_orm.extras if record_orm.extras else {},
            )

            task_specs[record_orm.id] = {
                "function": "qcengine.compute",
                "function_kwargs": {
                    "input_data": qcschema_input,
                    "program": specification.program,
                },
            }

        if set(record_ids) != set(task_specs.keys()):
            raise RuntimeError("Did not generate all task specs for all singlepoint records?")

        # Return in the input order
        return [task_specs[rid] for rid in record_ids]

    def create_wavefunction_orm(self, wavefunction: WavefunctionProperties) -> WavefunctionORM:
        """
        Convert a QCElemental wavefunction into a wavefunction ORM
        """

        wfn_dict = wavefunction.dict(encoding="json")
        cdata, ctype, clevel = compress(wfn_dict, CompressionEnum.zstd)

        return WavefunctionORM(compression_type=ctype, compression_level=clevel, data=cdata)

    def update_completed_schema_v1(self, session: Session, record_id: int, result: QCEl_AtomicResult) -> None:
        # Update the fields themselves
        if result.wavefunction:
            wavefunction_orm = self.create_wavefunction_orm(result.wavefunction)
            wavefunction_orm.record_id = record_id
            session.add(wavefunction_orm)

    def insert_full_qcportal_records_v1(
        self,
        session: Session,
        records: Sequence[SinglepointRecord],
        creator_user_id: Optional[int],
    ) -> List[SinglepointRecordORM]:

        # No child records, so creator_user_id is ignored
        ret = []

        mols = []
        qc_specs = []

        for record in records:
            mols.append(record.molecule)
            qc_specs.append(record.specification)

        meta, spec_ids = self.root_socket.records.singlepoint.add_specifications(qc_specs, session=session)
        if not meta.success:
            raise RuntimeError("Aborted single point insertion - could not add specifications: " + meta.error_string)

        meta, mol_ids = self.root_socket.molecules.add(mols, session=session)
        if not meta.success:
            raise RuntimeError("Aborted single point insertion - could not add molecules: " + meta.error_string)

        for record, mol_id, spec_id in zip(records, mol_ids, spec_ids):
            record_orm = SinglepointRecordORM(
                specification_id=spec_id,
                molecule_id=mol_id,
                status=RecordStatusEnum.complete,
            )

            if record.wavefunction_:
                wfn_orm = WavefunctionORM(
                    compression_type=record.wavefunction_.compression_type,
                    compression_level=-1,  # We don't know :( But it's not really used anywhere - just for curiosity
                    data=record.wavefunction_.data_,
                )
                record_orm.wavefunction = wfn_orm

            ret.append(record_orm)

        return ret

    def insert_full_schema_v1(
        self, session: Session, results: Sequence[QCEl_AtomicResult], creator_user_id: Optional[int]
    ) -> List[SinglepointRecordORM]:

        # No child records, so creator_user_id is ignored

        ret = []

        mols = []
        qc_specs = []

        for result in results:
            mols.append(result.molecule)

            qc_spec = QCSpecification(
                program=result.provenance.creator.lower(),
                driver=result.driver,
                method=result.model.method,
                basis=result.model.basis,
                keywords=result.keywords,
                protocols=result.protocols,
            )
            qc_specs.append(qc_spec)

        meta, spec_ids = self.root_socket.records.singlepoint.add_specifications(qc_specs, session=session)
        if not meta.success:
            raise RuntimeError("Aborted single point insertion - could not add specifications: " + meta.error_string)

        meta, mol_ids = self.root_socket.molecules.add(mols, session=session)
        if not meta.success:
            raise RuntimeError("Aborted single point insertion - could not add molecules: " + meta.error_string)

        for result, mol_id, spec_id in zip(results, mol_ids, spec_ids):
            record_orm = SinglepointRecordORM(
                specification_id=spec_id,
                molecule_id=mol_id,
                status=RecordStatusEnum.complete,
            )

            if result.wavefunction:
                record_orm.wavefunction = self.create_wavefunction_orm(result.wavefunction)

            ret.append(record_orm)

        return ret

    def add_specifications(
        self, qc_specs: Sequence[QCSpecification], *, session: Optional[Session] = None
    ) -> Tuple[InsertMetadata, List[int]]:
        """
        Adds specifications for singlepoint calculations to the database, returning their IDs.

        If an identical specification exists, then no insertion takes place and the id of the existing
        specification is returned.

        Specification IDs are returned in the same order as the input specifications

        Parameters
        ----------
        qc_specs
            Sequence of specifications to add to the database
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and the IDs of the specifications.
        """

        to_add = []

        for qc_spec in qc_specs:
            protocols_dict = qc_spec.protocols.dict(exclude_defaults=True, exclude_unset=True)

            # TODO - if error_correction is manually specified as the default, then it will be an empty dict
            if "error_correction" in protocols_dict:
                erc = protocols_dict["error_correction"]
                pol = erc.get("policies", dict())
                if len(pol) == 0:
                    erc.pop("policies", None)
                if len(erc) == 0:
                    protocols_dict.pop("error_correction")

            qc_spec_dict = qc_spec.dict(exclude={"basis", "protocols"})
            qc_spec_dict["basis"] = "" if qc_spec.basis is None else qc_spec.basis
            qc_spec_dict["protocols"] = protocols_dict
            qc_spec_hash = hash_dict(qc_spec_dict)

            qc_spec_orm = QCSpecificationORM(
                specification_hash=qc_spec_hash,
                program=qc_spec_dict["program"],
                driver=qc_spec_dict["driver"],
                method=qc_spec_dict["method"],
                basis=qc_spec_dict["basis"],
                keywords=qc_spec_dict["keywords"],
                protocols=qc_spec_dict["protocols"],
            )

            to_add.append(qc_spec_orm)

        with self.root_socket.optional_session(session, False) as session:
            meta, ids = insert_general(
                session,
                to_add,
                (QCSpecificationORM.specification_hash,),
                (QCSpecificationORM.id,),
                use_unique=True,
            )

            return meta, [x[0] for x in ids]

    def add_specification(
        self, qc_spec: QCSpecification, *, session: Optional[Session] = None
    ) -> Tuple[InsertMetadata, Optional[int]]:
        """
        Adds a single specification for a singlepoint calculation to the database, returning its id.

        If an identical specification exists, then no insertion takes place and the id of the existing
        specification is returned.

        Parameters
        ----------
        qc_spec
            Specification to add to the database
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and the id of the specification.
        """

        meta, ids = self.add_specifications([qc_spec], session=session)

        if not ids:
            return meta, None

        return meta, ids[0]

    def get(
        self,
        record_ids: Sequence[int],
        include: Optional[Sequence[str]] = None,
        exclude: Optional[Sequence[str]] = None,
        missing_ok: bool = False,
        *,
        session: Optional[Session] = None,
    ) -> List[Optional[Dict[str, Any]]]:
        options = []
        if include:
            if is_included("molecule", include, exclude, False) or is_included("molecules", include, exclude, False):
                options.append(joinedload(SinglepointRecordORM.molecule))
            if is_included("wavefunction", include, exclude, False):
                options.append(joinedload(SinglepointRecordORM.wavefunction).undefer(WavefunctionORM.data))

        with self.root_socket.optional_session(session, True) as session:
            return self.root_socket.records.get_base(
                orm_type=self.record_orm,
                record_ids=record_ids,
                include=include,
                exclude=exclude,
                missing_ok=missing_ok,
                additional_options=options,
                session=session,
            )

    def query(
        self,
        query_data: SinglepointQueryFilters,
        *,
        session: Optional[Session] = None,
    ) -> List[int]:
        and_query = []
        need_join = False

        if query_data.program is not None:
            and_query.append(QCSpecificationORM.program.in_(query_data.program))
            need_join = True
        if query_data.driver is not None:
            and_query.append(QCSpecificationORM.driver.in_(query_data.driver))
            need_join = True
        if query_data.method is not None:
            and_query.append(QCSpecificationORM.method.in_(query_data.method))
            need_join = True
        if query_data.basis is not None:
            and_query.append(QCSpecificationORM.basis.in_(query_data.basis))
            need_join = True
        if query_data.molecule_id is not None:
            and_query.append(SinglepointRecordORM.molecule_id.in_(query_data.molecule_id))
        if query_data.keywords is not None:
            or_query = []
            for d in query_data.keywords:
                or_query.append(QCSpecificationORM.keywords.comparator.contains(d))
            and_query.append(or_(*or_query))
            need_join = True

        stmt = select(SinglepointRecordORM.id)

        if need_join:
            stmt = stmt.join(SinglepointRecordORM.specification)

        stmt = stmt.where(*and_query)

        return self.root_socket.records.query_base(
            stmt=stmt,
            orm_type=SinglepointRecordORM,
            query_data=query_data,
            session=session,
        )

    def add_internal(
        self,
        molecule_ids: Sequence[int],
        qc_spec_id: int,
        compute_tag: str,
        compute_priority: PriorityEnum,
        creator_user_id: Optional[int],
        find_existing: bool,
        *,
        session: Optional[Session] = None,
    ) -> Tuple[InsertMetadata, List[Optional[int]]]:
        """
        Internal function for adding new singlepoint computations

        This function expects that the molecules and specification are already added to the
        database and that the ids are known.

        This checks if the calculations already exist in the database. If so, it returns
        the existing id, otherwise it will insert it and return the new id.

        Parameters
        ----------
        molecule_ids
            IDs of the molecules to run the computation with. One record will be added per molecule.
        qc_spec_id
            ID of the specification
        compute_tag
            The tag for the task. This will assist in routing to appropriate compute managers.
        compute_priority
            The priority for the computation
        creator_user_id
            ID of the user who created the record
        find_existing
            If True, search for existing records and return those. If False, always add new records
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and a list of record ids. The ids will be in the
            order of the input molecules
        """

        compute_tag = compute_tag.lower()

        with self.root_socket.optional_session(session, False) as session:
            # Get the spec orm. The full orm will be needed for create_task
            stmt = select(QCSpecificationORM).where(QCSpecificationORM.id == qc_spec_id)
            spec_orm = session.execute(stmt).scalar_one()

            all_orm = []

            for mid in molecule_ids:
                sp_orm = SinglepointRecordORM(
                    is_service=False,
                    specification=spec_orm,
                    specification_id=qc_spec_id,
                    molecule_id=mid,
                    status=RecordStatusEnum.waiting,
                    creator_user_id=creator_user_id,
                )

                self.create_task(sp_orm, compute_tag, compute_priority)
                all_orm.append(sp_orm)

            if find_existing:
                meta, ids = insert_general(
                    session,
                    all_orm,
                    (SinglepointRecordORM.specification_id, SinglepointRecordORM.molecule_id),
                    (SinglepointRecordORM.id,),
                    use_unique=False,
                    lock_id=singlepoint_insert_lock_id,
                )
                return meta, [x[0] for x in ids]
            else:
                session.add_all(all_orm)
                session.flush()
                meta = InsertMetadata(inserted_idx=list(range(len(all_orm))))

                return meta, [x.id for x in all_orm]

    def add(
        self,
        molecules: Sequence[Union[int, Molecule]],
        qc_spec: QCSpecification,
        compute_tag: str,
        compute_priority: PriorityEnum,
        creator_user: Optional[Union[int, str]],
        find_existing: bool,
        *,
        session: Optional[Session] = None,
    ) -> Tuple[InsertMetadata, List[Optional[int]]]:
        """
        Adds new singlepoint calculations

        This checks if the calculations already exist in the database. If so, it returns
        the existing id, otherwise it will insert it and return the new id.

        Parameters
        ----------
        molecules
            Molecules to compute using the specification
        qc_spec
            Specification for the calculations
        compute_tag
            The tag for the task. This will assist in routing to appropriate compute managers.
        compute_priority
            The priority for the computation
        creator_user
            Name or ID of the user who created the record
        find_existing
            If True, search for existing records and return those. If False, always add new records
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and a list of record ids. The ids will be in the
            order of the input molecules
        """

        with self.root_socket.optional_session(session, False) as session:
            creator_user_id = self.root_socket.users.get_optional_user_id(creator_user, session=session)

            # First, add the specification
            spec_meta, spec_id = self.add_specification(qc_spec, session=session)
            if not spec_meta.success:
                return (
                    InsertMetadata(
                        error_description="Aborted - could not add specification: " + spec_meta.error_string
                    ),
                    [],
                )

            # Now the molecules
            mol_meta, mol_ids = self.root_socket.molecules.add_mixed(molecules, session=session)
            if not mol_meta.success:
                return (
                    InsertMetadata(error_description="Aborted - could not add all molecules: " + mol_meta.error_string),
                    [],
                )

            return self.add_internal(
                mol_ids,
                spec_id,
                compute_tag,
                compute_priority,
                creator_user_id,
                find_existing,
                session=session,
            )

    def add_from_input(
        self,
        record_input: SinglepointInput,
        compute_tag: str,
        compute_priority: PriorityEnum,
        creator_user: Optional[Union[int, str]],
        find_existing: bool,
        *,
        session: Optional[Session] = None,
    ) -> Tuple[InsertCountsMetadata, int]:

        assert isinstance(record_input, SinglepointInput)

        meta, ids = self.add(
            [record_input.molecule],
            record_input.specification,
            compute_tag,
            compute_priority,
            creator_user,
            find_existing,
        )

        return InsertCountsMetadata.from_insert_metadata(meta), ids[0]

    ####################################################
    # Some stuff to be retrieved for singlepoints
    ####################################################

    def get_wavefunction_metadata(
        self, record_id: int, *, session: Optional[Session] = None
    ) -> Optional[Dict[str, Any]]:
        options = [
            lazyload("*"),
            defer("*"),
            joinedload(SinglepointRecordORM.wavefunction).options(
                undefer("*"), defaultload("*"), defer(WavefunctionORM.data)
            ),
        ]

        with self.root_socket.optional_session(session) as session:
            rec = session.get(SinglepointRecordORM, record_id, options=options)
            if rec is None:
                raise MissingDataError(f"Cannot find record {record_id}")
            if rec.wavefunction is None:
                return None
            return rec.wavefunction.model_dict()

    def get_wavefunction_rawdata(
        self, record_id: int, *, session: Optional[Session] = None
    ) -> Tuple[bytes, CompressionEnum]:
        options = [
            lazyload("*"),
            defer("*"),
            joinedload(SinglepointRecordORM.wavefunction).options(
                undefer(WavefunctionORM.data), undefer(WavefunctionORM.compression_type)
            ),
        ]

        with self.root_socket.optional_session(session) as session:
            rec = session.get(SinglepointRecordORM, record_id, options=options)
            if rec is None:
                raise MissingDataError(f"Cannot find record {record_id}")
            return rec.wavefunction.data, rec.wavefunction.compression_type



from __future__ import annotations

from sqlalchemy import Column, Integer, ForeignKey, String, ForeignKeyConstraint, Index, select
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import relationship

from qcfractal.components.dataset_db_models import BaseDatasetORM
from qcfractal.components.gridoptimization.record_db_models import (
    GridoptimizationRecordORM,
    GridoptimizationSpecificationORM,
)
from qcfractal.components.molecules.db_models import MoleculeORM
from qcfractal.db_socket import BaseORM


class GridoptimizationDatasetEntryORM(BaseORM):
    __tablename__ = "gridoptimization_dataset_entry"

    dataset_id = Column(Integer, ForeignKey("gridoptimization_dataset.id", ondelete="cascade"), primary_key=True)

    name = Column(String, primary_key=True)
    comment = Column(String)

    initial_molecule_id = Column(Integer, ForeignKey(MoleculeORM.id), nullable=False)
    additional_keywords = Column(JSONB, nullable=False)
    additional_optimization_keywords = Column(JSONB, nullable=False)
    attributes = Column(JSONB, nullable=False)

    initial_molecule = relationship(MoleculeORM, lazy="joined")

    __table_args__ = (
        Index("ix_gridoptimization_dataset_entry_dataset_id", "dataset_id"),
        Index("ix_gridoptimization_dataset_entry_name", "name"),
        Index("ix_gridoptimization_dataset_entry_initial_molecule_id", "initial_molecule_id"),
    )

    _qcportal_model_excludes = ["dataset_id", "initial_molecule_id"]


class GridoptimizationDatasetSpecificationORM(BaseORM):
    __tablename__ = "gridoptimization_dataset_specification"

    dataset_id = Column(Integer, ForeignKey("gridoptimization_dataset.id", ondelete="cascade"), primary_key=True)
    name = Column(String, primary_key=True)
    description = Column(String, nullable=True)
    specification_id = Column(Integer, ForeignKey(GridoptimizationSpecificationORM.id), nullable=False)

    specification = relationship(GridoptimizationSpecificationORM)

    __table_args__ = (
        Index("ix_gridoptimization_dataset_specification_dataset_id", "dataset_id"),
        Index("ix_gridoptimization_dataset_specification_name", "name"),
        Index("ix_gridoptimization_dataset_specification_specification_id", "specification_id"),
    )

    _qcportal_model_excludes = ["dataset_id", "specification_id"]


class GridoptimizationDatasetRecordItemORM(BaseORM):
    __tablename__ = "gridoptimization_dataset_record"

    dataset_id = Column(Integer, ForeignKey("gridoptimization_dataset.id", ondelete="cascade"), primary_key=True)
    entry_name = Column(String, primary_key=True)
    specification_name = Column(String, primary_key=True)
    record_id = Column(Integer, ForeignKey(GridoptimizationRecordORM.id), nullable=False)

    record = relationship(GridoptimizationRecordORM)

    __table_args__ = (
        ForeignKeyConstraint(
            ["dataset_id", "entry_name"],
            ["gridoptimization_dataset_entry.dataset_id", "gridoptimization_dataset_entry.name"],
            ondelete="cascade",
            onupdate="cascade",
        ),
        ForeignKeyConstraint(
            ["dataset_id", "specification_name"],
            ["gridoptimization_dataset_specification.dataset_id", "gridoptimization_dataset_specification.name"],
            ondelete="cascade",
            onupdate="cascade",
        ),
        Index("ix_gridoptimization_dataset_record_record_id", "record_id"),
    )

    _qcportal_model_excludes = ["dataset_id"]


class GridoptimizationDatasetORM(BaseDatasetORM):
    __tablename__ = "gridoptimization_dataset"

    id = Column(Integer, ForeignKey(BaseDatasetORM.id, ondelete="cascade"), primary_key=True)

    __mapper_args__ = {
        "polymorphic_identity": "gridoptimization",
    }


dataset_records_select = [
    select(
        BaseDatasetORM.id.label("dataset_id"),
        GridoptimizationDatasetRecordItemORM.entry_name.label("entry_name"),
        GridoptimizationDatasetRecordItemORM.specification_name.label("specification_name"),
        GridoptimizationDatasetRecordItemORM.record_id.label("record_id"),
    )
    .join(GridoptimizationDatasetRecordItemORM, BaseDatasetORM.id == GridoptimizationDatasetRecordItemORM.dataset_id)
    .where(BaseDatasetORM.dataset_type == "gridoptimization")
]



import requests

from qcarchivetesting.testing_classes import QCATestingSnowflake


def test_homepage_redirect(postgres_server, pytestconfig):
    pg_harness = postgres_server.get_new_harness("test_homepage_redirect")
    encoding = pytestconfig.getoption("--client-encoding")

    extra_config = {"homepage_redirect_url": "https://example.com"}
    snowflake = QCATestingSnowflake(pg_harness, encoding, extra_config=extra_config)

    # The URI doesn't contain any path or anything
    r = requests.get(snowflake.get_uri(), allow_redirects=False)
    assert r.status_code == 302
    assert r.headers["Location"] == "https://example.com"


def test_homepage_serve_dir(postgres_server, tmp_path, pytestconfig):
    encoding = pytestconfig.getoption("--client-encoding")

    homepage_dir = tmp_path / "homepage"
    homepage_dir.mkdir()

    index_file = str(homepage_dir / "index.html")
    other_file = str(homepage_dir / "other_file.html")

    with open(index_file, "w") as f:
        f.write("HOMEPAGE")
    with open(other_file, "w") as f:
        f.write("OTHER FILE")

    pg_harness = postgres_server.get_new_harness("test_homepage_serve_dir")
    extra_config = {"homepage_directory": str(homepage_dir)}
    snowflake = QCATestingSnowflake(pg_harness, encoding, extra_config=extra_config)

    # The URI doesn't contain any path or anything
    r = requests.get(snowflake.get_uri(), headers={"Accept": "text/html"})
    assert r.status_code == 200
    assert r.text == "HOMEPAGE"

    r = requests.get(snowflake.get_uri() + "/other_file.html", headers={"Accept": "text/html"})
    assert r.status_code == 200
    assert r.text == "OTHER FILE"

    r = requests.get(snowflake.get_uri() + "/missing.html", headers={"Accept": "text/html"})
    assert r.status_code == 404



"""Remove manager_name foreign key

Revision ID: 35bb042920a3
Revises: 0c0b1dea52dc
Create Date: 2025-07-09 14:17:24.601731

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "35bb042920a3"
down_revision = "0c0b1dea52dc"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(op.f("base_record_manager_name_fkey"), "base_record", type_="foreignkey")
    op.drop_constraint(op.f("record_compute_history_manager_name_fkey"), "record_compute_history", type_="foreignkey")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_foreign_key(
        op.f("record_compute_history_manager_name_fkey"),
        "record_compute_history",
        "compute_manager",
        ["manager_name"],
        ["name"],
    )
    op.create_foreign_key(
        op.f("base_record_manager_name_fkey"), "base_record", "compute_manager", ["manager_name"], ["name"]
    )
    # ### end Alembic commands ###



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from qcelemental.models import (
    OptimizationResult as QCEl_OptimizationResult,
)
from sqlalchemy import select, update
from sqlalchemy.orm import lazyload, joinedload, selectinload, defer, undefer, load_only

from qcfractal.components.singlepoint.record_db_models import QCSpecificationORM
from qcfractal.db_socket.helpers import insert_general
from qcportal.exceptions import MissingDataError
from qcportal.metadata_models import InsertMetadata, InsertCountsMetadata
from qcportal.molecules import Molecule
from qcportal.optimization import (
    OptimizationRecord,
    OptimizationSpecification,
    OptimizationQueryFilters,
    OptimizationInput,
    OptimizationMultiInput,
)
from qcportal.record_models import PriorityEnum, RecordStatusEnum
from qcportal.serialization import convert_numpy_recursive
from qcportal.singlepoint import QCSpecification
from qcportal.singlepoint import (
    SinglepointDriver,
)
from qcportal.utils import hash_dict, is_included
from .record_db_models import OptimizationSpecificationORM, OptimizationRecordORM, OptimizationTrajectoryORM
from ..base_record_socket import BaseRecordSocket

if TYPE_CHECKING:
    from sqlalchemy.orm.session import Session
    from qcfractal.db_socket.socket import SQLAlchemySocket
    from typing import List, Dict, Tuple, Optional, Sequence, Any, Union

# Meaningless, but unique to optimizations
optimization_insert_lock_id = 14100


class OptimizationRecordSocket(BaseRecordSocket):
    """
    Socket for handling optimization computations
    """

    # Used by the base class
    record_orm = OptimizationRecordORM
    record_input_type = OptimizationInput
    record_multi_input_type = OptimizationMultiInput

    def __init__(self, root_socket: SQLAlchemySocket):
        BaseRecordSocket.__init__(self, root_socket)
        self._logger = logging.getLogger(__name__)

    def generate_task_specifications(self, session: Session, record_ids: Sequence[int]) -> List[Dict[str, Any]]:

        stmt = select(OptimizationRecordORM).filter(OptimizationRecordORM.id.in_(record_ids))
        stmt = stmt.options(load_only(OptimizationRecordORM.id, OptimizationRecordORM.extras))
        stmt = stmt.options(
            lazyload("*"),
            joinedload(OptimizationRecordORM.initial_molecule),
            selectinload(OptimizationRecordORM.specification),
        )

        record_orms = session.execute(stmt).scalars().all()

        task_specs = {}

        for record_orm in record_orms:
            specification = record_orm.specification
            initial_molecule = record_orm.initial_molecule.model_dict()

            model = {"method": specification.qc_specification.method}

            # Empty basis string should be None in the model
            if specification.qc_specification.basis:
                model["basis"] = specification.qc_specification.basis
            else:
                model["basis"] = None

            # Add the singlepoint program to the optimization keywords
            opt_keywords = specification.keywords.copy()
            opt_keywords["program"] = specification.qc_specification.program

            # driver = "gradient" is what the current schema uses
            qcschema_input = dict(
                schema_name="qcschema_optimization_input",
                schema_version=1,
                id=str(record_orm.id),  # str for compatibility
                input_specification=dict(
                    schema_name="qcschema_input",
                    schema_version=1,
                    model=model,
                    driver="gradient",
                    keywords=specification.qc_specification.keywords,
                ),
                initial_molecule=convert_numpy_recursive(
                    initial_molecule, flatten=True
                ),  # TODO - remove after all data is converted
                keywords=opt_keywords,
                protocols=specification.protocols,
                extras=record_orm.extras if record_orm.extras else {},
            )

            # Note that the 'program' that runs an optimization is
            # called a 'procedure' in QCEngine
            task_specs[record_orm.id] = {
                "function": "qcengine.compute_procedure",
                "function_kwargs": {
                    "input_data": qcschema_input,
                    "procedure": specification.program,
                },
            }

        if set(record_ids) != set(task_specs.keys()):
            raise RuntimeError("Did not generate all task specs for all optimizaiton records?")

        # Return in the input order
        return [task_specs[rid] for rid in record_ids]

    def update_completed_schema_v1(self, session: Session, record_id: int, result: QCEl_OptimizationResult) -> None:

        # Add the final molecule
        meta, final_mol_id = self.root_socket.molecules.add([result.final_molecule], session=session)
        if not meta.success:
            raise RuntimeError("Unable to add final molecule: " + meta.error_string)

        # Insert the trajectory
        # Get the ID of the parent_record
        stmt = select(OptimizationRecordORM.creator_user_id).where(OptimizationRecordORM.id == record_id)
        creator_user_id = session.execute(stmt).scalar_one()

        traj_ids = self.root_socket.records.insert_full_schema_v1(session, result.trajectory, creator_user_id)

        for position, traj_id in enumerate(traj_ids):
            assoc_orm = OptimizationTrajectoryORM(singlepoint_id=traj_id)
            assoc_orm.optimization_id = record_id
            assoc_orm.position = position
            session.add(assoc_orm)

        # Update the fields themselves
        record_updates = {
            "final_molecule_id": final_mol_id[0],
            "energies": result.energies,
        }

        stmt = update(OptimizationRecordORM).where(OptimizationRecordORM.id == record_id).values(record_updates)
        session.execute(stmt)

    def insert_full_qcportal_records_v1(
        self,
        session: Session,
        records: Sequence[OptimizationRecord],
        creator_user_id: Optional[int],
    ) -> List[OptimizationRecordORM]:
        ret = []

        initial_mols = []
        final_mols = []
        opt_specs = []

        for record in records:
            initial_mols.append(record.initial_molecule)
            final_mols.append(record.final_molecule)
            opt_specs.append(record.specification)

        meta, spec_ids = self.root_socket.records.optimization.add_specifications(opt_specs, session=session)
        if not meta.success:
            raise RuntimeError("Aborted optimization insertion - could not add specifications: " + meta.error_string)

        meta, initial_mol_ids = self.root_socket.molecules.add(initial_mols, session=session)
        if not meta.success:
            raise RuntimeError("Aborted optimization insertion - could not add initial molecules: " + meta.error_string)

        for record, initial_mol_id, spec_id in zip(records, initial_mol_ids, spec_ids):

            final_mol_id = None
            if record.final_molecule:
                meta, final_mol_ids = self.root_socket.molecules.add([record.final_molecule], session=session)
                if not meta.success:
                    raise RuntimeError(
                        "Aborted optimization insertion - could not add final molecule: " + meta.error_string
                    )
                final_mol_id = final_mol_ids[0]

            record_orm = OptimizationRecordORM(
                specification_id=spec_id,
                initial_molecule_id=initial_mol_id,
                final_molecule_id=final_mol_id,
                energies=record.energies,
                status=RecordStatusEnum.complete,
            )

            if record.trajectory:
                trajectory_ids = self.root_socket.records.insert_full_qcportal_records(
                    session, record.trajectory, creator_user_id
                )
                opt_traj_orm = [
                    OptimizationTrajectoryORM(singlepoint_id=tid, position=idx)
                    for idx, tid in enumerate(trajectory_ids)
                ]
                record_orm.trajectory = opt_traj_orm

            ret.append(record_orm)

        return ret

    def insert_full_schema_v1(
        self,
        session: Session,
        results: Sequence[QCEl_OptimizationResult],
        creator_user_id: Optional[int],
    ) -> List[OptimizationRecordORM]:

        ret = []

        initial_mols = []
        final_mols = []
        opt_specs = []

        for result in results:
            initial_mols.append(result.initial_molecule)
            final_mols.append(result.final_molecule)

            # in v1 of the schema, the qc program is stored as a keyword
            qc_program = result.keywords.pop("program", "")

            qc_spec = QCSpecification(
                program=qc_program,
                driver=result.input_specification.driver,
                method=result.input_specification.model.method,
                basis=result.input_specification.model.basis,
                keywords=result.input_specification.keywords,
                # no protocols allowed in v1 of the schema
            )

            opt_spec = OptimizationSpecification(
                program=result.provenance.creator.lower(),
                qc_specification=qc_spec,
                keywords=result.keywords,
                protocols=result.protocols,
            )

            opt_specs.append(opt_spec)

        meta, spec_ids = self.root_socket.records.optimization.add_specifications(opt_specs, session=session)
        if not meta.success:
            raise RuntimeError("Aborted optimization insertion - could not add specifications: " + meta.error_string)

        meta, initial_mol_ids = self.root_socket.molecules.add(initial_mols, session=session)
        if not meta.success:
            raise RuntimeError("Aborted optimization insertion - could not add initial molecules: " + meta.error_string)

        meta, final_mol_ids = self.root_socket.molecules.add(final_mols, session=session)
        if not meta.success:
            raise RuntimeError("Aborted optimization insertion - could not add final molecules: " + meta.error_string)

        for result, initial_mol_id, final_mol_id, spec_id in zip(results, initial_mol_ids, final_mol_ids, spec_ids):
            record_orm = OptimizationRecordORM(
                specification_id=spec_id,
                initial_molecule_id=initial_mol_id,
                final_molecule_id=final_mol_id,
                energies=result.energies,
                status=RecordStatusEnum.complete,
            )

            if result.trajectory:
                trajectory_ids = self.root_socket.records.insert_full_schema_v1(
                    session, result.trajectory, creator_user_id
                )
                record_orm.trajectory = [
                    OptimizationTrajectoryORM(singlepoint_id=tid, position=idx)
                    for idx, tid in enumerate(trajectory_ids)
                ]

            ret.append(record_orm)

        return ret

    def add_specifications(
        self, opt_specs: Sequence[OptimizationSpecification], *, session: Optional[Session] = None
    ) -> Tuple[InsertMetadata, List[int]]:
        """
        Adds specification for optimization calculations to the database, returning their IDs.

        If an identical specification exists, then no insertion takes place and the id of the existing
        specification is returned.

        Specification IDs are returned in the same order as the input specifications

        Parameters
        ----------
        opt_specs
            Sequence of specification to add to the database
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and the IDs of the specifications.
        """

        to_add = []

        for opt_spec in opt_specs:
            protocols_dict = opt_spec.protocols.dict(exclude_defaults=True, exclude_unset=True)

            # Don't include lower specifications in the hash
            opt_spec_dict = opt_spec.dict(exclude={"protocols", "qc_specification"})
            opt_spec_dict["protocols"] = protocols_dict
            opt_spec_hash = hash_dict(opt_spec_dict)

            # Leave qc spec id for later
            opt_spec_orm = OptimizationSpecificationORM(
                program=opt_spec.program,
                keywords=opt_spec.keywords,
                protocols=protocols_dict,
                specification_hash=opt_spec_hash,
            )

            to_add.append(opt_spec_orm)

        with self.root_socket.optional_session(session, False) as session:
            qc_specs = [x.qc_specification for x in opt_specs]
            for qc_spec in qc_specs:
                # Make double sure the driver is deferred
                qc_spec.driver = SinglepointDriver.deferred

            meta, qc_spec_ids = self.root_socket.records.singlepoint.add_specifications(qc_specs, session=session)

            if not meta.success:
                return (
                    InsertMetadata(error_description="Unable to add single point specifications: " + meta.error_string),
                    [],
                )

            assert len(qc_spec_ids) == len(opt_specs)
            for opt_spec_orm, qc_spec_id in zip(to_add, qc_spec_ids):
                opt_spec_orm.qc_specification_id = qc_spec_id

            meta, ids = insert_general(
                session,
                to_add,
                (OptimizationSpecificationORM.specification_hash, OptimizationSpecificationORM.qc_specification_id),
                (OptimizationSpecificationORM.id,),
                use_unique=True,
            )

            return meta, [x[0] for x in ids]

    def add_specification(
        self, opt_spec: OptimizationSpecification, *, session: Optional[Session] = None
    ) -> Tuple[InsertMetadata, Optional[int]]:
        """
        Adds a specification for an optimization calculation to the database, returning its id.

        If an identical specification exists, then no insertion takes place and the id of the existing
        specification is returned.

        Parameters
        ----------
        opt_spec
            Specification to add to the database
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and the id of the specification.
        """

        meta, ids = self.add_specifications([opt_spec], session=session)

        if not ids:
            return meta, None

        return meta, ids[0]

    def get(
        self,
        record_ids: Sequence[int],
        include: Optional[Sequence[str]] = None,
        exclude: Optional[Sequence[str]] = None,
        missing_ok: bool = False,
        *,
        session: Optional[Session] = None,
    ) -> List[Optional[Dict[str, Any]]]:
        options = []
        if include:
            if is_included("initial_molecule", include, exclude, False) or is_included(
                "molecules", include, exclude, False
            ):
                options.append(joinedload(OptimizationRecordORM.initial_molecule))
            if is_included("final_molecule", include, exclude, False) or is_included(
                "molecules", include, exclude, False
            ):
                options.append(joinedload(OptimizationRecordORM.final_molecule))
            if is_included("trajectory", include, exclude, False):
                options.append(selectinload(OptimizationRecordORM.trajectory))

        with self.root_socket.optional_session(session, True) as session:
            return self.root_socket.records.get_base(
                orm_type=self.record_orm,
                record_ids=record_ids,
                include=include,
                exclude=exclude,
                missing_ok=missing_ok,
                additional_options=options,
                session=session,
            )

    def query(
        self,
        query_data: OptimizationQueryFilters,
        *,
        session: Optional[Session] = None,
    ) -> List[int]:
        and_query = []
        need_spspec_join = False
        need_optspec_join = False

        if query_data.program is not None:
            and_query.append(OptimizationSpecificationORM.program.in_(query_data.program))
            need_optspec_join = True
        if query_data.qc_program is not None:
            and_query.append(QCSpecificationORM.program.in_(query_data.qc_program))
            need_spspec_join = True
        if query_data.qc_method is not None:
            and_query.append(QCSpecificationORM.method.in_(query_data.qc_method))
            need_spspec_join = True
        if query_data.qc_basis is not None:
            and_query.append(QCSpecificationORM.basis.in_(query_data.qc_basis))
            need_spspec_join = True
        if query_data.initial_molecule_id is not None:
            and_query.append(OptimizationRecordORM.initial_molecule_id.in_(query_data.initial_molecule_id))
        if query_data.final_molecule_id is not None:
            and_query.append(OptimizationRecordORM.final_molecule_id.in_(query_data.final_molecule_id))

        stmt = select(OptimizationRecordORM.id)

        # If we need the singlepoint spec, we also need the optimization spec
        if need_optspec_join or need_spspec_join:
            stmt = stmt.join(OptimizationRecordORM.specification)

        if need_spspec_join:
            stmt = stmt.join(OptimizationSpecificationORM.qc_specification)

        stmt = stmt.where(*and_query)

        return self.root_socket.records.query_base(
            stmt=stmt,
            orm_type=OptimizationRecordORM,
            query_data=query_data,
            session=session,
        )

    def add_internal(
        self,
        initial_molecule_ids: Sequence[int],
        opt_spec_id: int,
        compute_tag: str,
        compute_priority: PriorityEnum,
        creator_user_id: Optional[int],
        find_existing: bool,
        *,
        session: Optional[Session] = None,
    ) -> Tuple[InsertMetadata, List[Optional[int]]]:
        """
        Internal function for adding new optimization computations

        This function expects that the molecules and specification are already added to the
        database and that the ids are known.

        This checks if the calculations already exist in the database. If so, it returns
        the existing id, otherwise it will insert it and return the new id.

        Parameters
        ----------
        initial_molecule_ids
            IDs of the molecules to optimize. One record will be added per molecule.
        opt_spec_id
            ID of the specification
        compute_tag
            The tag for the task. This will assist in routing to appropriate compute managers.
        compute_priority
            The priority for the computation
        creator_user_id
            ID of the user who created the record
        find_existing
            If True, search for existing records and return those. If False, always add new records
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and a list of record ids. The ids will be in the
            order of the input molecules
        """

        compute_tag = compute_tag.lower()

        with self.root_socket.optional_session(session) as session:
            # Get the spec orm. The full orm will be needed for create_task
            stmt = select(OptimizationSpecificationORM).where(OptimizationSpecificationORM.id == opt_spec_id)
            spec_orm = session.execute(stmt).scalar_one()

            all_orm = []
            all_molecules = self.root_socket.molecules.get(initial_molecule_ids, session=session)

            for mol_data in all_molecules:
                opt_orm = OptimizationRecordORM(
                    is_service=False,
                    specification=spec_orm,
                    specification_id=opt_spec_id,
                    initial_molecule_id=mol_data["id"],
                    status=RecordStatusEnum.waiting,
                    creator_user_id=creator_user_id,
                )

                self.create_task(opt_orm, compute_tag, compute_priority)
                all_orm.append(opt_orm)

            if find_existing:
                meta, ids = insert_general(
                    session,
                    all_orm,
                    (OptimizationRecordORM.specification_id, OptimizationRecordORM.initial_molecule_id),
                    (OptimizationRecordORM.id,),
                    use_unique=False,
                    lock_id=optimization_insert_lock_id,
                )
                return meta, [x[0] for x in ids]
            else:
                session.add_all(all_orm)
                session.flush()
                meta = InsertMetadata(inserted_idx=list(range(len(all_orm))))

                return meta, [x.id for x in all_orm]

    def add(
        self,
        initial_molecules: Sequence[Union[int, Molecule]],
        opt_spec: OptimizationSpecification,
        compute_tag: str,
        compute_priority: PriorityEnum,
        creator_user: Optional[Union[int, str]],
        find_existing: bool,
        *,
        session: Optional[Session] = None,
    ) -> Tuple[InsertMetadata, List[Optional[int]]]:
        """
        Adds new optimization calculations

        This checks if the calculations already exist in the database. If so, it returns
        the existing id, otherwise it will insert it and return the new id.

        Parameters
        ----------
        initial_molecules
            Molecules to compute using the specification
        opt_spec
            Specification for the calculations
        compute_tag
            The tag for the task. This will assist in routing to appropriate compute managers.
        compute_priority
            The priority for the computation
        creator_user
            Name or ID of the user who created the record
        find_existing
            If True, search for existing records and return those. If False, always add new records
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and a list of record ids. The ids will be in the
            order of the input molecules
        """

        with self.root_socket.optional_session(session, False) as session:
            creator_user_id = self.root_socket.users.get_optional_user_id(creator_user, session=session)

            # First, add the specification
            spec_meta, spec_id = self.add_specification(opt_spec, session=session)
            if not spec_meta.success:
                return (
                    InsertMetadata(
                        error_description="Aborted - could not add specification: " + spec_meta.error_string
                    ),
                    [],
                )

            # Now the molecules
            mol_meta, mol_ids = self.root_socket.molecules.add_mixed(initial_molecules, session=session)
            if not mol_meta.success:
                return (
                    InsertMetadata(error_description="Aborted - could not add all molecules: " + mol_meta.error_string),
                    [],
                )

            return self.add_internal(
                mol_ids,
                spec_id,
                compute_tag,
                compute_priority,
                creator_user_id,
                find_existing,
                session=session,
            )

    def add_from_input(
        self,
        record_input: OptimizationInput,
        compute_tag: str,
        compute_priority: PriorityEnum,
        creator_user: Optional[Union[int, str]],
        find_existing: bool,
        *,
        session: Optional[Session] = None,
    ) -> Tuple[InsertCountsMetadata, int]:

        assert isinstance(record_input, OptimizationInput)

        meta, ids = self.add(
            [record_input.initial_molecule],
            record_input.specification,
            compute_tag,
            compute_priority,
            creator_user,
            find_existing,
        )

        return InsertCountsMetadata.from_insert_metadata(meta), ids[0]

    ####################################################
    # Some stuff to be retrieved for optimizations
    ####################################################

    def get_trajectory_ids(
        self,
        record_id: int,
        *,
        session: Optional[Session] = None,
    ) -> List[int]:
        """
        Retrieve the IDs of the singlepoint computations that form the trajectory
        """

        options = [
            lazyload("*"),
            defer("*"),
            joinedload(OptimizationRecordORM.trajectory).options(undefer(OptimizationTrajectoryORM.singlepoint_id)),
        ]

        with self.root_socket.optional_session(session) as session:
            rec = session.get(OptimizationRecordORM, record_id, options=options)
            if rec is None:
                raise MissingDataError(f"Cannot find record {record_id}")
            return [x.singlepoint_id for x in rec.trajectory]



from __future__ import annotations

import socket
import threading
import time
from datetime import timedelta
from typing import TYPE_CHECKING

import pytest

from qcarchivetesting import test_users
from qcfractal.components.internal_jobs.socket import InternalJobSocket
from qcportal import PortalRequestError
from qcportal.internal_jobs import InternalJobStatusEnum
from qcportal.utils import now_at_utc

if TYPE_CHECKING:
    from qcarchivetesting.testing_classes import QCATestingSnowflake


# Add in another function to the internal_jobs socket for testing
def dummmy_internal_job(self, iterations: int, session, job_progress):
    for i in range(iterations):
        time.sleep(1.0)
        job_progress.update_progress(100 * ((i + 1) / iterations), f"Interation {i} of {iterations}")
        # print("Dummy internal job counter ", i)

        job_progress.raise_if_cancelled()

    return "Internal job finished"


# And another one for errors
def dummmy_internal_job_error(self, session, job_progress):
    raise RuntimeError("Expected error")


setattr(InternalJobSocket, "client_dummy_job", dummmy_internal_job)
setattr(InternalJobSocket, "client_dummy_job_error", dummmy_internal_job_error)


def test_internal_jobs_client_error(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    id_1 = storage_socket.internal_jobs.add(
        "client_dummy_job_error", now_at_utc(), "internal_jobs.client_dummy_job_error", {}, None, unique_name=False
    )

    # Faster updates for testing
    storage_socket.internal_jobs._update_frequency = 1

    time_0 = now_at_utc()
    end_event = threading.Event()
    th = threading.Thread(target=storage_socket.internal_jobs.run_loop, args=(end_event,))
    th.start()
    time.sleep(3)
    time_1 = now_at_utc()
    end_event.set()
    th.join()

    job_1 = snowflake_client.get_internal_job(id_1)
    assert job_1.status == InternalJobStatusEnum.error
    assert job_1.progress < 100
    assert time_0 < job_1.ended_date < time_1
    assert time_0 < job_1.last_updated < time_1
    assert "Expected error" in job_1.result


def test_internal_jobs_client_cancel_waiting(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    id_1 = storage_socket.internal_jobs.add(
        "client_dummy_job", now_at_utc(), "internal_jobs.client.dummy_job", {"iterations": 10}, None, unique_name=False
    )

    snowflake_client.cancel_internal_job(id_1)

    job_1 = snowflake_client.get_internal_job(id_1)
    assert job_1.status == InternalJobStatusEnum.cancelled
    assert job_1.result is None
    assert job_1.started_date is None
    assert job_1.progress == 0


def test_internal_jobs_client_cancel_running(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    id_1 = storage_socket.internal_jobs.add(
        "client_dummy_job", now_at_utc(), "internal_jobs.client_dummy_job", {"iterations": 10}, None, unique_name=False
    )

    # Faster updates for testing
    storage_socket.internal_jobs._update_frequency = 1

    end_event = threading.Event()
    th = threading.Thread(target=storage_socket.internal_jobs.run_loop, args=(end_event,))
    th.start()
    time.sleep(4)

    try:
        job_1 = snowflake_client.get_internal_job(id_1)
        assert job_1.status == InternalJobStatusEnum.running
        assert job_1.progress > 10

        snowflake_client.cancel_internal_job(id_1)
        time.sleep(6)

        job_1 = snowflake_client.get_internal_job(id_1)
        assert job_1.status == InternalJobStatusEnum.cancelled
        assert job_1.progress < 70
        assert job_1.result is None

    finally:
        end_event.set()
        th.join()


def test_internal_jobs_client_delete_waiting(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    id_1 = storage_socket.internal_jobs.add(
        "client_dummy_job", now_at_utc(), "internal_jobs.client_dummy_job", {"iterations": 10}, None, unique_name=False
    )

    snowflake_client.delete_internal_job(id_1)

    with pytest.raises(PortalRequestError, match="Internal job.*not found"):
        snowflake_client.get_internal_job(id_1)


def test_internal_jobs_client_delete_running(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    id_1 = storage_socket.internal_jobs.add(
        "client_dummy_job", now_at_utc(), "internal_jobs.client_dummy_job", {"iterations": 10}, None, unique_name=False
    )

    # Faster updates for testing
    storage_socket.internal_jobs._update_frequency = 1

    end_event = threading.Event()
    th = threading.Thread(target=storage_socket.internal_jobs.run_loop, args=(end_event,))
    th.start()
    time.sleep(4)

    try:
        job_1 = snowflake_client.get_internal_job(id_1)
        assert job_1.status == InternalJobStatusEnum.running
        assert job_1.progress > 10

        snowflake_client.delete_internal_job(id_1)
        time.sleep(6)

        with pytest.raises(PortalRequestError, match="Internal job.*not found"):
            snowflake_client.get_internal_job(id_1)

    finally:
        end_event.set()
        th.join()


def test_internal_jobs_client_query(secure_snowflake: QCATestingSnowflake):
    client = secure_snowflake.client("admin_user", test_users["admin_user"]["pw"])
    storage_socket = secure_snowflake.get_storage_socket()

    read_id = client.get_user("read_user").id

    time_0 = now_at_utc()
    id_1 = storage_socket.internal_jobs.add(
        "client_dummy_job",
        now_at_utc(),
        "internal_jobs.client_dummy_job",
        {"iterations": 1},
        read_id,
        unique_name=False,
    )
    time_1 = now_at_utc()

    # Faster updates for testing
    storage_socket.internal_jobs._update_frequency = 1

    end_event = threading.Event()
    th = threading.Thread(target=storage_socket.internal_jobs.run_loop, args=(end_event,))
    th.start()
    time.sleep(4)
    time_2 = now_at_utc()

    try:
        job_1 = client.get_internal_job(id_1)
        assert job_1.status == InternalJobStatusEnum.complete

    finally:
        end_event.set()
        th.join()

    # Add one that will be waiting
    id_2 = storage_socket.internal_jobs.add(
        "client_dummy_job", now_at_utc(), "internal_jobs.client_dummy_job", {"iterations": 1}, None, unique_name=False
    )

    time_3 = now_at_utc()

    # Now do some queries
    result = client.query_internal_jobs(job_id=id_1)
    r = list(result)
    assert len(r) == 1
    assert r[0].id == id_1

    result = client.query_internal_jobs(user="read_user")
    r = list(result)
    assert len(r) == 1
    assert r[0].id == id_1

    result = client.query_internal_jobs(user=read_id)
    r = list(result)
    assert len(r) == 1
    assert r[0].id == id_1

    result = client.query_internal_jobs(name="client_dummy_job", status=["complete"])
    r = list(result)
    assert len(r) == 1
    assert r[0].id == id_1

    result = client.query_internal_jobs(name="client_dummy_job", status="waiting")
    r = list(result)
    assert len(r) == 1
    assert r[0].id == id_2

    result = client.query_internal_jobs(name="client_dummy_job", added_after=time_0)
    r = list(result)
    assert len(r) == 2
    assert {r[0].id, r[1].id} == {id_1, id_2}

    result = client.query_internal_jobs(name="client_dummy_job", added_after=time_1, added_before=time_3)
    r = list(result)
    assert len(r) == 1
    assert r[0].id == id_2

    result = client.query_internal_jobs(name="client_dummy_job", last_updated_after=time_2)
    r = list(result)
    assert len(r) == 0

    result = client.query_internal_jobs(name="client_dummy_job", last_updated_before=time_2)
    r = list(result)
    assert len(r) == 1
    assert r[0].id == id_1

    result = client.query_internal_jobs(runner_hostname=socket.gethostname())
    result_l = list(result)
    assert len(result_l) >= 2

    # Empty query - all
    # This should be at least two - other services were probably added by the socket
    result = client.query_internal_jobs()
    result_l = list(result)
    assert len(result_l) >= 2

    # Should also have a bunch
    result = client.query_internal_jobs(scheduled_before=now_at_utc())
    result_l = list(result)
    assert len(result_l) >= 2

    # nothing scheduled 10 days from now
    result = client.query_internal_jobs(scheduled_after=now_at_utc() + timedelta(days=10))
    result_l = list(result)
    assert len(result_l) == 0

    # Some queries for non-existent stuff
    result = client.query_internal_jobs(name="abcd")
    result_l = list(result)
    assert len(result_l) == 0

    # Some queries for non-existent stuff
    result = client.query_internal_jobs(status="error")
    assert len(result_l) == 0



"""add more info to project tables

Revision ID: 4d5ad63ad478
Revises: 0390661b91bf
Create Date: 2025-04-21 15:56:32.589003

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "4d5ad63ad478"
down_revision = "0390661b91bf"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("project_dataset", sa.Column("name", sa.String(), nullable=False))
    op.add_column(
        "project_dataset",
        sa.Column(
            "lname",
            sa.String(length=100),
            sa.Computed(
                "LOWER(name)",
            ),
            nullable=False,
        ),
    )
    op.add_column("project_dataset", sa.Column("description", sa.String(), nullable=False))
    op.add_column("project_dataset", sa.Column("tagline", sa.String(), nullable=False))
    op.add_column("project_dataset", sa.Column("tags", sa.JSON(), nullable=False))

    op.drop_constraint("project_dataset_pkey", "project_dataset", type_="primary")
    op.create_primary_key(
        "project_dataset_pkey",
        "project_dataset",
        ["project_id", "lname"],
    )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    raise NotImplementedError("Cannot downgrade")
    # ### end Alembic commands ###



import json
import os
import tarfile
import zipfile
from typing import Iterable, List, Tuple, Dict

from qcportal.molecules import Molecule, MoleculeUploadOptions


def _molecule_from_file(fname, data: str) -> Molecule:
    # Split into a separate function for future expansion
    if fname.lower().endswith(".json"):
        mol_data = json.loads(data)
        return Molecule(**mol_data)
    else:
        return Molecule.from_data(data)


def molecules_from_files(
    file_list: Iterable[Tuple[str, str]], options: MoleculeUploadOptions
) -> Tuple[Dict[str, List[Tuple[str, Molecule]]], List[str]]:
    """
    Load molecules from a list of files, supporting archives.

    Parameters
    ----------
    file_list
        Iterable of (filename, file_path). Filename includes extension to determine type.
    options
        Upload options (currently unused, reserved for future behavior tweaks).

    Returns
    -------
    :
        A tuple. The first part is a list of errors found when processing the files.
        The second part is a dictionary of filenames to list of molecule info. The molecule info is a tuple
        containing the inner filename and a Molecule object.

    Notes
    -----
    - Archives are processed in-memory; no extraction to filesystem occurs.
    - Directories within archives are ignored (no recursion).
    - Supported archives: .zip, .tar, .tar.gz, .tgz, .tar.bz2, .tbz2, .tar.xz, .txz
    """

    errors: List[str] = []
    results: Dict[str, List[Tuple[str, Molecule]]] = {}

    for fname, fpath in file_list:
        results[fname] = []

        lower_fname = fname.lower()

        # ZIP archives
        if lower_fname.endswith(".zip"):
            try:
                with zipfile.ZipFile(fpath) as zf:
                    for info in zf.infolist():
                        # Skip directories
                        if info.is_dir():
                            continue
                        data = zf.read(info.filename)
                        inner_name = os.path.basename(info.filename)

                        try:
                            mol = _molecule_from_file(inner_name, data.decode())
                            results[fname].append((inner_name, mol))
                        except Exception as e:
                            errors.append(f"Error parsing molecule '{fname} : {inner_name}': {e}")

            except Exception as e:
                errors.append(f"Error decompressing ZIP archive '{fname}': {e}")

            continue

        # TAR archives (auto-detect compression)
        elif lower_fname.endswith((".tar", ".tar.gz", ".tgz", ".tar.bz2", ".tbz2", ".tar.xz", ".txz")):
            try:
                # r:* autodetects compression (gz, bz2, xz) if available
                with tarfile.open(fpath, mode="r:*") as tf:
                    for member in tf.getmembers():
                        if not member.isfile():
                            continue
                        extracted = tf.extractfile(member)
                        if extracted is None:
                            continue
                        with extracted:
                            data = extracted.read()

                        inner_name = os.path.basename(member.name) or member.name

                        try:
                            mol = _molecule_from_file(inner_name, data.decode())
                            results[fname].append((inner_name, mol))
                        except Exception as e:
                            errors.append(f"Error parsing molecule '{fname} : {inner_name}': {e}")

            except Exception as e:
                errors.append(f"Error decompressing ZIP archive '{fname}': {e}")

            continue

        # Plain molecule file
        else:
            with open(fpath, "rt") as fh:
                data = fh.read()

                try:
                    mol = _molecule_from_file(os.path.basename(fpath), data)
                    results[fname].append((fname, mol))
                except Exception as e:
                    errors.append(f"Error parsing molecule '{fname}': {e}")

    return results, errors



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Tuple, List, Any, Optional

from sqlalchemy import select, delete
from sqlalchemy.dialects.postgresql import insert

from qcportal.auth import UserInfo
from qcportal.exceptions import AuthenticationFailure, AuthorizationFailure
from qcportal.utils import now_at_utc
from .db_models import UserORM, UserSessionORM
from .permission_evaluation import evaluate_global_permissions

if TYPE_CHECKING:
    import datetime
    from sqlalchemy.orm.session import Session
    from qcfractal.db_socket.socket import SQLAlchemySocket


class AuthSocket:
    """
    Socket for authenticating and authorizing
    """

    def __init__(self, root_socket: SQLAlchemySocket):
        self.root_socket = root_socket
        self._logger = logging.getLogger(__name__)

        self.security_enabled = self.root_socket.qcf_config.enable_security
        self.allow_unauthenticated_read = self.root_socket.qcf_config.allow_unauthenticated_read

    def verify(self, user_id: int, *, session: Optional[Session] = None) -> UserInfo:
        """
        Verifies that a given user id exists and is enabled, returning info about the user and their role

        This does not check the user's password.

        If the user is not found, or is disabled, an exception is raised.

        Parameters
        ----------
        user_id
            The id of the user to check
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        --------
        :
            All information about the user, and all information about the user's role
        """

        stmt = select(UserORM)
        stmt = stmt.where(UserORM.id == user_id)

        with self.root_socket.optional_session(session, True) as session:
            user_orm: Optional[UserORM] = session.execute(stmt).scalar_one_or_none()

            if user_orm is None:
                raise AuthenticationFailure("User does not exist")

            if not user_orm.enabled:
                raise AuthenticationFailure(f"User {user_id} is disabled.")

            user_info = user_orm.to_model(UserInfo)
            return user_info

    def assert_global_permission(self, role: Optional[str], resource: str, action: str, require_security: bool):
        # Some endpoints require security to be enabled
        if not self.security_enabled:
            if require_security:
                raise AuthorizationFailure(f"Cannot access '{resource}' with security disabled")
            else:
                return

        # Use anonymous role if no role is given
        if role is None:
            role = "anonymous"

        # Don't allow the anonymous role unless the server is set up to allow it
        if role == "anonymous" and not self.allow_unauthenticated_read:
            raise AuthenticationFailure("Server requires login")

        allowed = evaluate_global_permissions(role, resource, action)

        if not allowed:
            raise AuthorizationFailure(
                f"Role '{role}' is not authorized to use action '{action}' on resource '{resource}'"
            )

    def allowed_actions(self, subject: Any, resources: Any, actions: Any, policies: Any) -> List[Tuple[str, str]]:
        raise NotImplementedError("TODO")

    def save_user_session(
        self,
        user_id: int,
        user_session_key: str,
        user_session_data: Any,
        *,
        session: Optional[Session] = None,
    ) -> None:
        """
        Saves user/flask session data to the database
        """

        with self.root_socket.optional_session(session, False) as session:
            stmt = insert(UserSessionORM)
            stmt = stmt.values(user_id=user_id, session_key=user_session_key, session_data=user_session_data)
            stmt = stmt.on_conflict_do_update(
                index_elements=[UserSessionORM.session_key],
                set_={"session_data": user_session_data, "last_accessed": now_at_utc()},
            )
            session.execute(stmt)

    def load_user_session(
        self, user_session_key: str, *, session: Optional[Session] = None
    ) -> Tuple[Any, datetime.datetime]:
        """
        Loads user/flask session data from the database

        Will return None if the session_key does not exist in the database

        Returns a tuple of the session data and the last accessed time
        """
        with self.root_socket.optional_session(session, True) as session:
            stmt = select(UserSessionORM).where(UserSessionORM.session_key == user_session_key)
            flask_session_orm = session.execute(stmt).scalar_one_or_none()

            if not flask_session_orm:
                return None, now_at_utc()

            return flask_session_orm.session_data, flask_session_orm.last_accessed

    def delete_user_session(
        self,
        user_session_key: Optional[str] = None,
        user_session_public_id: Optional[int] = None,
        *,
        session: Optional[Session] = None,
    ) -> None:
        """
        Deletes user/flask session data from the database (if it exists)
        """

        if (user_session_key is None) and (user_session_public_id is None):
            raise ValueError("Either user_session_key or user_session_public_id (but not both) must be specified")
        if (user_session_key is not None) and (user_session_public_id is not None):
            raise ValueError("Either user_session_key or user_session_public_id (but not both) must be specified")

        with self.root_socket.optional_session(session, False) as session:
            stmt = delete(UserSessionORM)

            if user_session_key is not None:
                stmt = stmt.where(UserSessionORM.session_key == user_session_key)
            if user_session_public_id is not None:
                stmt = stmt.where(UserSessionORM.public_id == user_session_public_id)

            session.execute(stmt)

    def list_all_user_sessions(self, *, session: Optional[Session] = None) -> List[Tuple[int, datetime.datetime]]:
        """
        List all sessions currently in the database
        """

        with self.root_socket.optional_session(session, True) as session:
            stmt = select(UserSessionORM)
            session_orm = session.execute(stmt).scalars().all()
            return [s.public_dict() for s in session_orm]

    def list_user_sessions(
        self, user_id: int, *, session: Optional[Session] = None
    ) -> List[Tuple[int, datetime.datetime]]:
        """
        List all sessions currently in the database for a single user
        """

        with self.root_socket.optional_session(session, True) as session:
            stmt = select(UserSessionORM)
            stmt = stmt.where(UserSessionORM.user_id == user_id)
            session_orm = session.execute(stmt).scalars().all()
            return [s.public_dict() for s in session_orm]

    def clear_user_sessions(self, user_id: int, *, session: Optional[Session] = None):
        """
        Clear all sessions for a single user
        """

        with self.root_socket.optional_session(session) as session:
            stmt = delete(UserSessionORM)
            stmt = stmt.where(UserSessionORM.user_id == user_id)
            session.execute(stmt)



from __future__ import annotations

from typing import TYPE_CHECKING

from qcarchivetesting import load_molecule_data
from qcportal.molecules import Molecule, MoleculeQueryFilters

if TYPE_CHECKING:
    from qcfractal.db_socket import SQLAlchemySocket


def test_molecules_socket_validated_fix(storage_socket: SQLAlchemySocket):
    water = load_molecule_data("water_dimer_minima")
    hooh = load_molecule_data("hooh")

    # Add once
    meta, ids = storage_socket.molecules.add([water, hooh])

    mols = storage_socket.molecules.get(ids)
    assert mols[0]["validated"] is True
    assert mols[0]["fix_com"] is True
    assert mols[0]["fix_orientation"] is True
    assert mols[1]["validated"] is True
    assert mols[1]["fix_com"] is True
    assert mols[1]["fix_orientation"] is True


def test_molecules_socket_get_proj(storage_socket: SQLAlchemySocket):
    water = load_molecule_data("water_dimer_minima")
    hooh = load_molecule_data("hooh")

    # Add once
    meta, ids = storage_socket.molecules.add([water, hooh])

    mols = storage_socket.molecules.get(ids, include=["symbols", "geometry", "fix_com"])
    assert len(mols) == 2
    assert set(mols[0].keys()) == {"id", "symbols", "geometry", "fix_com"}
    assert set(mols[1].keys()) == {"id", "symbols", "geometry", "fix_com"}

    mols = storage_socket.molecules.get(ids, exclude=["symbols", "geometry", "fix_com"])
    assert len(mols) == 2
    assert set(mols[0].keys()).intersection({"symbols", "geometry", "fix_com"}) == set()
    assert set(mols[1].keys()).intersection({"symbols", "geometry", "fix_com"}) == set()

    mols = storage_socket.molecules.get(ids, include=["symbols", "geometry", "fix_com", "fix_orientation", "validated"])
    assert len(mols) == 2
    assert set(mols[0].keys()) == {"id", "symbols", "geometry", "fix_com", "fix_orientation", "validated"}
    assert set(mols[1].keys()) == {"id", "symbols", "geometry", "fix_com", "fix_orientation", "validated"}
    assert mols[0]["validated"] is True
    assert mols[1]["validated"] is True


def test_molecules_socket_add_mixed_1(storage_socket: SQLAlchemySocket):
    # Tests a simple add_mixed
    water = load_molecule_data("water_dimer_minima")
    hooh = load_molecule_data("hooh")
    ne4 = load_molecule_data("neon_tetramer")

    meta, ids = storage_socket.molecules.add_mixed([hooh, water, hooh, water, water, ne4])
    assert meta.success
    assert meta.n_inserted == 3
    assert meta.n_existing == 3
    assert ids[0] == ids[2]
    assert ids[1] == ids[3]
    assert ids[1] == ids[4]
    assert meta.inserted_idx == [0, 1, 5]
    assert meta.existing_idx == [2, 3, 4]

    meta, ids = storage_socket.molecules.add_mixed([ids[0], ids[0], ids[1], ids[1], ids[0]])
    assert meta.success
    assert meta.n_inserted == 0
    assert meta.n_existing == 5
    assert ids[0] == ids[1]
    assert ids[2] == ids[3]
    assert ids[0] == ids[4]


def test_molecules_socket_add_mixed_2(storage_socket: SQLAlchemySocket):
    # Tests a simple add_mixed
    water = load_molecule_data("water_dimer_minima")
    hooh = load_molecule_data("hooh")
    ne4 = load_molecule_data("neon_tetramer")

    meta, ids = storage_socket.molecules.add([water])
    assert meta.success

    meta, new_ids = storage_socket.molecules.add_mixed([hooh, ids[0]])
    assert meta.success
    assert meta.n_inserted == 1
    assert meta.n_existing == 1
    assert ids[0] == new_ids[1]

    # Try with duplicates, including a duplicate specified by molecule and by id
    meta, new_ids_2 = storage_socket.molecules.add_mixed([ne4, hooh, ids[0], new_ids[0], water, ids[0]])
    assert meta.success
    assert meta.n_inserted == 1
    assert meta.n_existing == 5
    assert meta.inserted_idx == [0]
    assert new_ids_2[1] == new_ids[0]
    assert new_ids_2[2] == ids[0]
    assert new_ids_2[3] == new_ids[0]
    assert new_ids_2[4] == ids[0]
    assert new_ids_2[5] == ids[0]

    mols = storage_socket.molecules.get([ids[0], new_ids[0], new_ids[1]], missing_ok=False)
    mols = [Molecule(**x) for x in mols]
    assert mols[0] == water
    assert mols[1] == hooh
    assert mols[2] == water


def test_molecules_socket_add_mixed_bad_1(storage_socket: SQLAlchemySocket):
    # Tests add_mixed but giving it a nonexistant id
    water = load_molecule_data("water_dimer_minima")
    hooh = load_molecule_data("hooh")

    meta, ids = storage_socket.molecules.add([water])
    assert meta.success

    meta, new_ids = storage_socket.molecules.add_mixed([hooh, 12345, 67890, water])
    assert not meta.success
    assert meta.n_inserted == 1
    assert meta.n_existing == 1
    assert len(meta.errors) == 2
    assert ids[0] == new_ids[3]
    assert meta.error_idx == [1, 2]
    assert "MoleculeORM object with id=12345 was not found" in meta.errors[0][1]
    assert "MoleculeORM object with id=67890 was not found" in meta.errors[1][1]



from __future__ import annotations

import itertools
from typing import TYPE_CHECKING, Optional

import pytest

from qcarchivetesting import load_molecule_data
from qcfractal.components.gridoptimization.record_db_models import GridoptimizationRecordORM
from qcportal.gridoptimization import GridoptimizationKeywords, GridoptimizationSpecification
from qcportal.optimization import OptimizationSpecification
from qcportal.record_models import RecordStatusEnum, PriorityEnum
from qcportal.singlepoint import QCSpecification
from qcportal.utils import now_at_utc
from .testing_helpers import compare_gridoptimization_specs, test_specs, submit_procedure_data, run_procedure_data

if TYPE_CHECKING:
    from qcarchivetesting.testing_classes import QCATestingSnowflake
    from qcportal import PortalClient


def test_gridoptimization_client_tag_priority(snowflake_client: PortalClient):
    peroxide2 = load_molecule_data("peroxide2")

    for tag, priority in itertools.product(["*", "tag99"], list(PriorityEnum)):
        # add tag/priority to keywords to force adding new record
        meta1, id1 = snowflake_client.add_gridoptimizations(
            [peroxide2],
            "gridoptimization",
            optimization_specification=OptimizationSpecification(
                program="geometric",
                qc_specification=QCSpecification(
                    program="psi4",
                    driver="deferred",
                    method="hf",
                    basis="sto-3g",
                    keywords={"tag_priority": [tag, priority]},
                ),
            ),
            keywords=GridoptimizationKeywords(
                preoptimization=False,
                scans=[
                    {"type": "distance", "indices": [1, 2], "steps": [-0.1, 0.0], "step_type": "relative"},
                    {"type": "dihedral", "indices": [0, 1, 2, 3], "steps": [-90, 0], "step_type": "absolute"},
                ],
            ),
            compute_priority=priority,
            compute_tag=tag,
        )

        assert meta1.n_inserted == 1
        rec = snowflake_client.get_records(id1, include=["service"])
        assert rec[0].service.compute_tag == tag
        assert rec[0].service.compute_priority == priority


@pytest.mark.parametrize("spec", test_specs)
def test_gridoptimization_client_add_get(
    submitter_client: PortalClient,
    spec: GridoptimizationSpecification,
):
    hooh = load_molecule_data("peroxide2")
    h3ns = load_molecule_data("go_H3NS")

    time_0 = now_at_utc()
    meta, id = submitter_client.add_gridoptimizations(
        [hooh, h3ns],
        spec.program,
        spec.optimization_specification,
        spec.keywords,
        compute_tag="tag1",
        compute_priority=PriorityEnum.low,
    )
    time_1 = now_at_utc()
    assert meta.success

    recs = submitter_client.get_gridoptimizations(id, include=["service", "initial_molecule"])
    assert len(recs) == 2

    for r in recs:
        assert r.record_type == "gridoptimization"
        assert r.record_type == "gridoptimization"
        assert compare_gridoptimization_specs(spec, r.specification)

        assert r.status == RecordStatusEnum.waiting
        assert r.children_status == {}

        assert r.service.compute_tag == "tag1"
        assert r.service.compute_priority == PriorityEnum.low

        assert r.creator_user == submitter_client.username

        assert time_0 < r.created_on < time_1
        assert time_0 < r.modified_on < time_1

    assert recs[0].initial_molecule.identifiers.molecule_hash == hooh.get_hash()
    assert recs[1].initial_molecule.identifiers.molecule_hash == h3ns.get_hash()


@pytest.mark.parametrize("spec", test_specs)
@pytest.mark.parametrize("find_existing", [True, False])
def test_gridoptimization_client_add_duplicate(
    submitter_client: PortalClient,
    spec: GridoptimizationSpecification,
    find_existing,
):
    hooh = load_molecule_data("peroxide2")
    h3ns = load_molecule_data("go_H3NS")
    all_mols = [hooh, h3ns]

    meta, id = submitter_client.add_gridoptimizations(
        all_mols,
        spec.program,
        spec.optimization_specification,
        spec.keywords,
        compute_tag="tag1",
        compute_priority=PriorityEnum.low,
        find_existing=True,
    )
    assert meta.success

    meta, id2 = submitter_client.add_gridoptimizations(
        all_mols,
        spec.program,
        spec.optimization_specification,
        spec.keywords,
        compute_tag="tag1",
        compute_priority=PriorityEnum.low,
        find_existing=find_existing,
    )
    assert meta.success

    if find_existing:
        assert meta.n_existing == len(all_mols)
        assert meta.n_inserted == 0
        assert id == id2
    else:
        assert meta.n_existing == 0
        assert meta.n_inserted == len(all_mols)
        assert set(id).isdisjoint(id2)


def test_gridoptimization_client_add_existing_molecule(snowflake_client: PortalClient):
    spec = test_specs[0]

    mol1 = load_molecule_data("go_H3NS")
    mol2 = load_molecule_data("peroxide2")

    # Add a molecule separately
    _, mol_ids = snowflake_client.add_molecules([mol2])

    # Now add records
    meta, id = snowflake_client.add_gridoptimizations(
        [mol1, mol2, mol2, mol1],
        "gridoptimization",
        keywords=spec.keywords,
        optimization_specification=spec.optimization_specification,
        compute_tag="tag1",
        compute_priority=PriorityEnum.low,
    )

    assert meta.success
    assert meta.n_inserted == 2
    assert meta.n_existing == 2

    recs = snowflake_client.get_gridoptimizations(id, include=["initial_molecule"])
    assert len(recs) == 4
    assert recs[0].id == recs[3].id
    assert recs[1].id == recs[2].id

    rec_mols = {x.initial_molecule.id for x in recs}
    _, mol_ids_2 = snowflake_client.add_molecules([mol1])
    assert rec_mols == set(mol_ids + mol_ids_2)


def test_gridoptimization_client_delete(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    go_id = run_procedure_data(storage_socket, activated_manager_name, "go_H2O2_psi4_pbe")

    with storage_socket.session_scope() as session:
        rec = session.get(GridoptimizationRecordORM, go_id)
        child_ids = [x.optimization_id for x in rec.optimizations]

    meta = snowflake_client.delete_records(go_id, soft_delete=True, delete_children=False)
    assert meta.success
    assert meta.deleted_idx == [0]
    assert meta.n_children_deleted == 0

    child_recs = snowflake_client.get_records(child_ids, missing_ok=True)
    assert all(x.status == RecordStatusEnum.complete for x in child_recs)
    go_rec = snowflake_client.get_records(go_id)
    assert go_rec.children_status == {RecordStatusEnum.complete: len(child_ids)}

    snowflake_client.undelete_records(go_id)

    meta = snowflake_client.delete_records(go_id, soft_delete=True, delete_children=True)
    assert meta.success
    assert meta.deleted_idx == [0]
    assert meta.n_children_deleted == len(child_ids)

    child_recs = snowflake_client.get_records(child_ids, missing_ok=True)
    assert all(x.status == RecordStatusEnum.deleted for x in child_recs)
    go_rec = snowflake_client.get_records(go_id)
    assert go_rec.children_status == {RecordStatusEnum.deleted: len(child_ids)}

    meta = snowflake_client.delete_records(go_id, soft_delete=False, delete_children=True)
    assert meta.success
    assert meta.deleted_idx == [0]
    assert meta.n_children_deleted == len(child_ids)

    recs = snowflake_client.get_gridoptimizations(go_id, missing_ok=True)
    assert recs is None

    child_recs = snowflake_client.get_records(child_ids, missing_ok=True)
    assert all(x is None for x in child_recs)

    # DB should be pretty empty now
    query_res = snowflake_client.query_records()
    query_res_l = list(query_res)
    assert len(query_res_l) == 0


def test_gridoptimization_client_harddelete_nochildren(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    go_id = run_procedure_data(storage_socket, activated_manager_name, "go_H2O2_psi4_pbe")

    with storage_socket.session_scope() as session:
        rec = session.get(GridoptimizationRecordORM, go_id)
        child_ids = [x.optimization_id for x in rec.optimizations]

    meta = snowflake_client.delete_records(go_id, soft_delete=False, delete_children=False)
    assert meta.success
    assert meta.deleted_idx == [0]
    assert meta.n_children_deleted == 0

    recs = snowflake_client.get_gridoptimizations(go_id, missing_ok=True)
    assert recs is None

    child_recs = snowflake_client.get_records(child_ids, missing_ok=True)
    assert all(x is not None for x in child_recs)


def test_gridoptimization_client_delete_opt_inuse(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    go_id = run_procedure_data(storage_socket, activated_manager_name, "go_H2O2_psi4_pbe")

    with storage_socket.session_scope() as session:
        rec = session.get(GridoptimizationRecordORM, go_id)
        child_ids = [x.optimization_id for x in rec.optimizations]

    meta = snowflake_client.delete_records(child_ids[0], soft_delete=False)
    assert meta.success is False
    assert meta.error_idx == [0]

    ch_rec = snowflake_client.get_records(child_ids[0])
    assert ch_rec is not None


def test_gridoptimization_client_query(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    id_1, _ = submit_procedure_data(storage_socket, "go_H2O2_psi4_b3lyp")
    id_2, _ = submit_procedure_data(storage_socket, "go_H2O2_psi4_pbe")
    id_3, _ = submit_procedure_data(storage_socket, "go_C4H4N2OS_mopac_pm6")
    id_4, _ = submit_procedure_data(storage_socket, "go_H3NS_psi4_pbe")

    all_gos = snowflake_client.get_gridoptimizations([id_1, id_2, id_3, id_4])
    mol_ids = [x.initial_molecule_id for x in all_gos]

    query_res = snowflake_client.query_gridoptimizations(qc_program=["psi4"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 3

    query_res = snowflake_client.query_gridoptimizations(qc_program=["nothing"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 0

    query_res = snowflake_client.query_gridoptimizations(initial_molecule_id=[mol_ids[0], 9999])
    query_res_l = list(query_res)
    assert len(query_res_l) == 2

    # query for optimization program
    query_res = snowflake_client.query_gridoptimizations(optimization_program=["geometric"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 4

    # query for optimization program
    query_res = snowflake_client.query_gridoptimizations(optimization_program=["geometric123"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 0

    # query for basis
    query_res = snowflake_client.query_gridoptimizations(qc_basis=["sTO-3g"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 3

    query_res = snowflake_client.query_gridoptimizations(qc_basis=[None])
    query_res_l = list(query_res)
    assert len(query_res_l) == 1

    query_res = snowflake_client.query_gridoptimizations(qc_basis=[""])
    query_res_l = list(query_res)
    assert len(query_res_l) == 1

    # query for method
    query_res = snowflake_client.query_gridoptimizations(qc_method=["b3lyP"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 1

    # Query by default returns everything
    query_res = snowflake_client.query_gridoptimizations()
    query_res_l = list(query_res)
    assert len(query_res_l) == 4

    # Query by default (with a limit)
    query_res = snowflake_client.query_gridoptimizations(limit=1)
    query_res_l = list(query_res)
    assert len(query_res_l) == 1



from __future__ import annotations

import pytest

from qcarchivetesting import test_users
from qcarchivetesting.testing_classes import QCATestingSnowflake
from qcfractal.components.serverinfo.test_access_socket import test_ips
from qcportal import PortalClient
from qcportal.utils import now_at_utc


@pytest.fixture(scope="module")
def queryable_access_client(postgres_server, pytestconfig):
    pg_harness = postgres_server.get_new_harness("serverinfo_test_access")
    encoding = pytestconfig.getoption("--client-encoding")

    # Don't log accesses
    with QCATestingSnowflake(pg_harness, encoding, enable_security=True, create_users=True, log_access=False) as server:
        # generate a bunch of test data
        storage_socket = server.get_storage_socket()

        admin_uid = storage_socket.users.get("admin_user")["id"]
        read_uid = storage_socket.users.get("read_user")["id"]

        with storage_socket.session_scope() as session:
            for i in range(10):
                for user_id in [admin_uid, read_uid]:
                    for endpoint in ["molecules", "records", "wavefunctions", "managers"]:
                        for method in ["GET", "POST"]:
                            access = {
                                "module": "api" if i % 2 == 0 else "compute",
                                "full_uri": f"api/v1/{endpoint}",
                                "method": method,
                                "ip_address": test_ips[0][0],
                                "user_agent": "Fake user agent",
                                "request_duration": 0.25 * i,
                                "user_id": user_id,
                                "request_bytes": 2 * i,
                                "response_bytes": 4 * i,
                            }
                            storage_socket.serverinfo.save_access(access, session=session)

        yield server.client("admin_user", test_users["admin_user"]["pw"])


def test_serverinfo_client_query_access(queryable_access_client: PortalClient):
    query_res = queryable_access_client.query_access_log(method=["get"])
    all_entries = list(query_res)
    assert len(all_entries) == 80

    query_res = queryable_access_client.query_access_log(method=["POST"])
    all_entries = list(query_res)
    assert len(all_entries) == 80

    query_res = queryable_access_client.query_access_log(module=["api"])
    all_entries = list(query_res)
    assert len(all_entries) == 80

    query_res = queryable_access_client.query_access_log(module=["compute"])
    all_entries = list(query_res)
    assert len(all_entries) == 80

    query_res = queryable_access_client.query_access_log(user="admin_user")
    all_entries = list(query_res)
    assert len(all_entries) == 80

    read_id = queryable_access_client.get_user("read_user").id
    query_res = queryable_access_client.query_access_log(user=[read_id, "admin_user"])
    all_entries = list(query_res)
    assert len(all_entries) == 160

    query_res = queryable_access_client.query_access_log(user=["no_user"])
    all_entries = list(query_res)
    assert len(all_entries) == 0


def test_serverinfo_client_query_access_empty_iter(queryable_access_client: PortalClient):
    query_res = queryable_access_client.query_access_log()
    assert len(query_res._current_batch) < queryable_access_client.api_limits["get_access_logs"]

    all_entries = list(query_res)
    assert len(all_entries) == 160


def test_serverinfo_client_query_access_limit(queryable_access_client: PortalClient):
    query_res = queryable_access_client.query_access_log(limit=99)
    all_entries = list(query_res)
    assert len(all_entries) == 99


def test_serverinfo_client_access_summary(queryable_access_client: PortalClient):
    # Just test that it works
    # TODO - better way of testing. Prepopulated db?
    now = now_at_utc()
    r = queryable_access_client.query_access_summary()
    assert list(r.entries.keys())[0] == now.strftime("%Y-%m-%d")

    r = queryable_access_client.query_access_summary(group_by="user")
    assert set(r.entries.keys()) == {"admin_user", "read_user"}



from __future__ import annotations

from typing import TYPE_CHECKING

from sqlalchemy import (
    Column,
    String,
    Integer,
    ForeignKey,
    Enum,
    TIMESTAMP,
    JSON,
    Index,
    Boolean,
    LargeBinary,
    UniqueConstraint,
    DDL,
    event,
)
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import relationship, deferred
from sqlalchemy.orm.collections import attribute_keyed_dict

from qcfractal.components.auth.db_models import UserORM, UserIDMapSubquery
from qcfractal.db_socket import BaseORM
from qcportal.compression import CompressionEnum, decompress
from qcportal.record_models import RecordStatusEnum, OutputTypeEnum
from qcportal.utils import now_at_utc

if TYPE_CHECKING:
    from typing import Dict, Any, Optional, Iterable


class RecordCommentORM(BaseORM):
    """
    Table for storing comments on calculations
    """

    __tablename__ = "record_comment"

    id = Column(Integer, primary_key=True)
    record_id = Column(Integer, ForeignKey("base_record.id", ondelete="cascade"), nullable=False)
    timestamp = Column(TIMESTAMP(timezone=True), default=now_at_utc, nullable=False)
    comment = Column(String, nullable=False)

    user_id = Column(Integer, ForeignKey(UserORM.id), nullable=True)
    user = relationship(
        UserIDMapSubquery,
        foreign_keys=[user_id],
        primaryjoin="RecordCommentORM.user_id == UserIDMapSubquery.id",
        lazy="selectin",
    )

    __table_args__ = (Index("ix_record_comment_record_id", "record_id"),)

    _qcportal_model_excludes = ["user_id", "user"]

    def model_dict(self, exclude: Optional[Iterable[str]] = None) -> Dict[str, Any]:
        d = BaseORM.model_dict(self, exclude)

        d["username"] = self.user.username if self.user is not None else None
        return d


class RecordInfoBackupORM(BaseORM):
    """
    Table for storing backup info about a record

    This stores previous tag, status, priority, etc, for a record. This is used when undoing
    delete, canceling, etc.
    """

    __tablename__ = "record_info_backup"

    id = Column(Integer, primary_key=True)
    record_id = Column(Integer, ForeignKey("base_record.id", ondelete="cascade"), nullable=False)
    old_status = Column(Enum(RecordStatusEnum), nullable=False)
    old_compute_tag = Column(String, nullable=True)
    old_compute_priority = Column(Integer, nullable=True)
    modified_on = Column(TIMESTAMP(timezone=True), default=now_at_utc, nullable=False)

    __table_args__ = (Index("ix_record_info_backup_record_id", "record_id"),)

    _qcportal_model_excludes = ["id", "record_id"]


class OutputStoreORM(BaseORM):
    """
    Table for storing raw computation outputs (text) and errors (json)
    """

    __tablename__ = "output_store"

    id = Column(Integer, primary_key=True)
    history_id = Column(Integer, ForeignKey("record_compute_history.id", ondelete="cascade"), nullable=False)

    output_type = Column(Enum(OutputTypeEnum), nullable=False)
    compression_type = Column(Enum(CompressionEnum), nullable=False)
    compression_level = Column(Integer, nullable=False)
    data = deferred(Column(LargeBinary, nullable=False))

    __table_args__ = (UniqueConstraint("history_id", "output_type", name="ux_output_store_id_type"),)

    _qcportal_model_excludes = ["id", "history_id", "compression_level"]

    def get_output(self) -> Any:
        return decompress(self.data, self.compression_type)


# Mark the storage of the data column as external
event.listen(
    OutputStoreORM.__table__,
    "after_create",
    DDL("ALTER TABLE output_store ALTER COLUMN data SET STORAGE EXTERNAL").execute_if(dialect=("postgresql")),
)


class NativeFileORM(BaseORM):
    """
    Table for storing raw, program-dependent raw data
    """

    __tablename__ = "native_file"

    id = Column(Integer, primary_key=True)
    record_id = Column(Integer, ForeignKey("base_record.id", ondelete="cascade"), nullable=False)

    name = Column(String, nullable=False)
    compression_type = Column(Enum(CompressionEnum), nullable=False)
    compression_level = Column(Integer, nullable=False)
    data = deferred(Column(LargeBinary, nullable=False))

    __table_args__ = (UniqueConstraint("record_id", "name", name="ux_native_file_record_id_name"),)

    _qcportal_model_excludes = ["id", "record_id", "compression_level"]

    def get_file(self) -> Any:
        return decompress(self.data, self.compression_type)


# Mark the storage of the data column of native files as external
event.listen(
    NativeFileORM.__table__,
    "after_create",
    DDL("ALTER TABLE native_file ALTER COLUMN data SET STORAGE EXTERNAL").execute_if(dialect=("postgresql")),
)


class RecordComputeHistoryORM(BaseORM):
    """
    Table for storing the computation history of a record

    The computation history stores the result status, provenance, and manager info that
    ran a computation. This is useful for storing the history of records that have errored multiple
    times.
    """

    __tablename__ = "record_compute_history"

    id = Column(Integer, primary_key=True)
    record_id = Column(Integer, ForeignKey("base_record.id", ondelete="cascade"), nullable=False)

    status = Column(Enum(RecordStatusEnum), nullable=False)
    manager_name = Column(String, nullable=True)
    modified_on = Column(TIMESTAMP(timezone=True), default=now_at_utc, nullable=False)
    provenance = Column(JSON)

    outputs = relationship(
        OutputStoreORM,
        collection_class=attribute_keyed_dict("output_type"),
        cascade="all, delete-orphan",
        passive_deletes=True,
    )

    __table_args__ = (
        Index("ix_record_compute_history_record_id", "record_id"),
        Index("ix_record_compute_history_manager_name", "manager_name"),
    )


class BaseRecordORM(BaseORM):
    """
    Base class for all the kinds of records
    """

    __tablename__ = "base_record"

    # for SQLAlchemy inheritence
    record_type = Column(String(100), nullable=False)

    # Some records can be either a service or a procedure
    is_service = Column(Boolean, nullable=False)

    # Base identification
    id = Column(Integer, primary_key=True)

    # Extra fields
    extras = Column(JSONB, nullable=False, default=dict)

    # Compute status
    # (Denormalized from compute history table for faster lookup during manager claiming/returning)
    status = Column(Enum(RecordStatusEnum), nullable=False)
    manager_name = Column(String, nullable=True)

    created_on = Column(TIMESTAMP(timezone=True), default=now_at_utc, nullable=False)
    modified_on = Column(TIMESTAMP(timezone=True), default=now_at_utc, nullable=False)

    # Who created this record
    creator_user_id = Column(Integer, ForeignKey(UserORM.id), nullable=True)

    creator_user = relationship(
        UserIDMapSubquery,
        foreign_keys=[creator_user_id],
        primaryjoin="BaseRecordORM.creator_user_id == UserIDMapSubquery.id",
        lazy="selectin",
        viewonly=True,
    )

    # Full compute history
    compute_history = relationship(
        RecordComputeHistoryORM,
        order_by=RecordComputeHistoryORM.modified_on.asc(),
        cascade="all, delete-orphan",
        passive_deletes=True,
    )

    comments = relationship(
        RecordCommentORM, order_by=RecordCommentORM.timestamp.asc(), cascade="all, delete-orphan", passive_deletes=True
    )

    # Related task. The foreign key is in the task_queue table
    task = relationship(
        "TaskQueueORM", back_populates="record", uselist=False, cascade="all, delete-orphan", passive_deletes=True
    )

    # Related service. The foreign key is in the service_queue table
    service = relationship(
        "ServiceQueueORM", back_populates="record", uselist=False, cascade="all, delete-orphan", passive_deletes=True
    )

    # Backed-up info (used for undelete, etc)
    info_backup = relationship(
        RecordInfoBackupORM,
        order_by=RecordInfoBackupORM.modified_on.asc(),
        cascade="all, delete-orphan",
        passive_deletes=True,
    )

    # Various computed properties and stuff
    properties = Column(JSONB)

    # Native files returned from the computation
    native_files = relationship(
        NativeFileORM, collection_class=attribute_keyed_dict("name"), cascade="all, delete-orphan", passive_deletes=True
    )

    __table_args__ = (
        Index("ix_base_record_status", "status"),
        Index("ix_base_record_record_type", "record_type"),
        Index("ix_base_record_manager_name", "manager_name"),
        Index("ix_base_record_creator_user_id", "creator_user_id"),
        Index("ix_base_record_created_on", "created_on"),
        Index("ix_base_record_modified_on", "modified_on"),
    )

    __mapper_args__ = {"polymorphic_on": "record_type"}

    # strip user/group ids
    # info_backup is also never part of models
    _qcportal_model_excludes = ["creator_user_id", "info_backup"]

    def model_dict(self, exclude: Optional[Iterable[str]] = None) -> Dict[str, Any]:
        d = BaseORM.model_dict(self, exclude)

        # TODO - DEPRECATED - REMOVE EVENTUALLY
        d["owner_user"] = self.creator_user.username if self.creator_user is not None else None
        d.pop("creator_user", None)
        d["owner_group"] = None

        return d

    @property
    def short_description(self) -> str:
        """
        Returns a short string describing the record
        """

        return f"short_description not implemented for {self.record_type}"


# Function for deleting large binary when derived classes are deleted
_del_baserecord_triggerfunc = DDL(
    """
    CREATE OR REPLACE FUNCTION public.qca_base_record_delete()
    RETURNS trigger
    LANGUAGE plpgsql
    AS $_$
        BEGIN
          DELETE FROM base_record WHERE base_record.id = OLD.id;
          RETURN OLD;
        END
        $_$
    ;
"""
)

event.listen(BaseRecordORM.__table__, "after_create", _del_baserecord_triggerfunc.execute_if(dialect=("postgresql")))



"""Merge claim_speed and nullable properties branches

Revision ID: d03466436fec
Revises: 12e2ba353ee6, 03c96181c90f
Create Date: 2024-12-06 11:32:14.436022

"""

# revision identifiers, used by Alembic.
revision = "d03466436fec"
down_revision = ("12e2ba353ee6", "03c96181c90f")
branch_labels = None
depends_on = None


def upgrade():
    # Nothing to do here. Branches were completely compatible
    pass


def downgrade():
    # Nothing to do here. Branches were completely compatible
    pass
